(* Copyright 2011 Cornell University
 *
 *
 * This file is part of EventML - a tool aiming at specifying
 * distributed protocols in an ML like language.  It is an interface
 * to the logic of events and is compiled into Nuprl.  It is written
 * by the NUPRL group of Cornell University, Ithaca, NY.
 *
 * EventML is a free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EventML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EventML.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  o Authors:     Vincent Rahli
 *  o Affiliation: Cornell University, NUPRL group
 *  o Date:        20 May 2011
 *  o File name:   Evaluator1.sml
 *  o Description: .
 *)


structure Evaluator1 = struct

structure B = Tools
structure T = NuprlTerms
structure P = Primitive
structure E = EvalProperties

 (* all 3 are used by TERMOF *)
fun mk_tag_term _ = raise Fail "tag_term"
fun TagC _ = raise Fail "TagC"
fun apply_conv _ = raise Fail "apply_conv"

fun compare_atomn n value1 value2 =
    let val op_id = T.opid_of_term value1
    in case (T.parameters_of_term value1,
	     T.parameters_of_term value2) of
	   ([param1], [param2]) =>
	   (* NOTE: both values should only have one parameter *)
	   let val ptype = T.type_of_parameter param1
	   in if op_id = "token"
		 andalso op_id = T.opid_of_term value2
		 (* NOTE: both values have to be tokens *)
		 andalso ptype = T.type_of_parameter param2
		 (* NOTE: both value have to have the the same kind of parameters *)
		 andalso (case n of (* NOTE: what are these: *)
			      0 => (ptype = "token" orelse ptype = "t")
			    | 1 => ptype = "ut1"
			    | 2 => ptype = "ut2"
			    | _ => false)
	      then T.equal_parameters param1 param2
	      else raise Fail "compare_atomn"
	   end
	 | _ => raise Fail "compare_atomn"
    end

val dummy_opid = "dummy"
val dummy_term = T.mk_nuprl_simple_term dummy_opid []

structure MAP = BinaryMapFn(type ord_key = string val compare = String.compare)

(* The nuprl part correspond to Nuprl's library and
 * the user part correspond to the user code.
 * In user, the binders go from left to right. *)
type item = string * T.nuprl_term * T.nuprl_term
type lib  = item MAP.map

val ab_ref = ref ("", dummy_term, dummy_term)

val found_user : T.nuprl_term option ref = ref NONE

fun set_found_user t = found_user := SOME t
fun get_found_user u =
    let val t = !found_user
	val _ = found_user := NONE
    in Option.getOpt (t, u)
    end

val emlib : lib = MAP.empty

fun union_libs (lib1 : lib) (lib2 : lib) =
    MAP.unionWithi (fn (id, x, y) => y) (lib1, lib2)

fun insert_lib lib id (obid, lhs, rhs) =
    MAP.insert (lib, id, (obid, lhs, rhs))

fun mk_ilibrary t u = T.mk_nuprl_simple_term "!library" [t, u]

fun strip_ilib term =
    let val ((opid, params), bterms) = T.dest_term term
    in if opid = "!library"
       then strip_ilib (#2 (hd bterms))
       else T.mk_nuprl_term (opid, params) (map (fn (vars, t) => (vars, strip_ilib t)) bterms)
    end

fun search_in_user user opid =
    case T.dest_simple_term user of
	(_, [lhs, rhs, tail]) =>
	let val id = T.opid_of_term lhs
	in if id = opid
	   then (ab_ref := ("", lhs, rhs);
		 set_found_user tail;
		 true)
	   else search_in_user tail opid
	end
      | _ => false

fun search_in_nuprl (libmap : lib) opid =
    case MAP.find (libmap, opid) of
	SOME (obid, lhs, rhs) => (ab_ref := (obid, lhs, rhs); true)
      | NONE => false

fun is_abstraction_term (libmap, user) term =
    let val opid = T.opid_of_term term
    in search_in_user user opid
       orelse
       search_in_nuprl libmap opid
    end

fun unfold_ab term =
    let val (obid, lhs, rhs) = !ab_ref
    in T.unfold_ab term lhs rhs
    end

fun NextStepEval1 t steps0 principals non_principals steps (lib as (map,user)) =
    let val opid = T.opid_of_term t

    in if B.member opid ["add", "subtract", "multiply", "divide", "remainder"]
       then let val [(v1,u1),(v2,u2)] = principals
	    in (P.do_primitive_int_op opid v1 v2, user, B.decr_steps steps, false)
	    end

       else if opid = "minus"
       then let val [(v,u)] = principals
	    in if T.is_nuprl_term "natural_number" v
	       then (T.mk_nuprl_simple_term "minus" [v], u, steps, false)
	       else (P.do_primitive_minus v, u, B.decr_steps steps, false)
	    end

       else if B.member opid ["less", "int_eq"]
       then let val [(v1,u1),(v2,u2)] = principals
		val [([],t3),([],t4)] = non_principals
	    in if P.do_primitive_cmp opid v1 v2
	       then (t3, user, B.decr_steps steps, true)
	       else (t4, user, B.decr_steps steps, true)
	    end

       else if opid = "atom_eq"
       then let val n = (T.firstnat t) handle _ => 0
		val [(v1,u1),(v2,u2)] = principals
		val [([],t3),([],t4)] = non_principals
	    in if compare_atomn n v1 v2
	       then (t3, user, B.decr_steps steps, true)
	       else (t4, user, B.decr_steps steps, true)
	    end

       else if opid = "eq_term"
       then let val [(v1,u1),(v2,u2)] = principals
	    in if P.is_complete_primitive_value v1
		  andalso
		  P.is_complete_primitive_value v2
	       then if T.alpha_equal_terms v1 v2
		    then (T.mk_nuprl_inl_term T.mk_nuprl_axiom_term, user, B.decr_steps steps, false)
		    else (T.mk_nuprl_inr_term T.mk_nuprl_axiom_term, user, B.decr_steps steps, false)
	       else raise Fail "eq_term"
	    end

       else if B.member opid ["isinr", "isinl", "ispair", "isint", "islambda", "isatom2"]
       then let val [(v1,u1)] = principals
		val [([],t2),([],t3)] = non_principals
	    in if P.do_primitive_test opid v1
	       then (t2, user, B.decr_steps steps, true)
	       else (t3, user, B.decr_steps steps, true)
	    end

       else if opid = "spread"
       then let val [(v1,u1)] = principals
		val [([x,y],B)] = non_principals
		val (a, b) = T.dest_pair v1
		val t1 = mk_ilibrary a u1
		val t2 = mk_ilibrary b u1
	    in (T.fo_subst [(x,t1),(y,t2)] B, user, B.decr_steps steps, true)
	    end

       else if opid = "decide"
       then let val [(v1,u1)] = principals
		val [([x],A),([y],B)] = non_principals
		val t1 = mk_ilibrary (T.subtermn 1 v1) u1
	    in if T.is_nuprl_term "inl" v1
	       then (T.fo_subst [(x, t1)] A, user, B.decr_steps steps, true)
	       else if T.is_nuprl_term "inr" v1
	       then (T.fo_subst [(y, t1)] B, user, B.decr_steps steps, true)
	       else raise Fail "decide"
	    end

       else if opid = "apply"
       then let val [([],arg)] = non_principals
	    in if steps < 0 andalso T.is_nuprl_term "ycomb" (T.subtermn 1 t)
	       then (T.mk_nuprl_apply_term arg t, user, B.decr_steps steps0, true)
	       else let val [(f,u)] = principals
			val (x, B) = T.dest_lambda f
			val targ = mk_ilibrary arg user
		    in (T.fo_subst [(x, targ)] B, u, B.decr_steps steps, true)
		    end
	    end

       else if opid = "callbyvalue"
       then let val [(q,u)] = principals
		val [([x],B)] = non_principals
		val tq = mk_ilibrary q u
	    in if P.is_primitive_value q
	       then (T.fo_subst [(x, tq)] B, user, B.decr_steps steps, true)
	       else raise Fail "callbyvalue"
	    end

       else if opid = "callbyvalueall"
       then let val [(q,u)] = principals
		val [([x],B)] = non_principals
		val tq = mk_ilibrary q u
	    in if P.is_complete_primitive_value q
	       then (T.fo_subst [(x, tq)] B, user, B.decr_steps steps, true)
	       else raise Fail "callbyvalueall"
	    end
(**)
       else if opid = "list_ind"
       then let val [(q,u)] = principals
		val [([],nilcase),([x,xs,r],conscase)] = non_principals
	    in if T.is_nuprl_term "nil" q
	       then (nilcase, user, B.decr_steps steps, true)
	       else if T.is_nuprl_term "cons" q
	       then let val [qa,qb] = T.subterms q
			val t2 = T.mk_nuprl_list_ind_term qb nilcase (x,xs,r,conscase)
		    in (T.fo_subst [(x,qa),(xs,qb),(r,t2)] conscase, user, B.decr_steps steps, true)
		    end
	       else raise Fail "list_ind"
	    end

       else if opid = "ind"
       then let val [(q,u)] = principals
		val [([x,rd],downcase),([],basecase),([y,ru],upcase)] = non_principals
		val n = T.dest_integer q
		val t' = if n = 0 then basecase
			 else let val (m,r,w,c) =
				      if n > 0
				      then (n-1,ru,y,upcase)
				      else (n+1,rd,x,downcase)
				  val p = T.mk_nuprl_integer_term m
				  val t2 = T.mk_nuprl_ind_term p (x,rd,downcase) basecase (y,ru,upcase)
			      in (T.fo_subst [(w,q),(r,t2)] c)
			      end
	    in (t', user, B.decr_steps steps, true)
	    end

       else if opid = "rec_ind"
       then let val [([],arg),([f,x],B)] = non_principals
		val B' = T.mk_nuprl_rec_ind_term (T.mk_nuprl_variable_term x) (f,x,B)
		val function = T.mk_nuprl_lambda_term x B'
	    in (T.fo_subst [(x,arg),(f,function)] B, user, B.decr_steps steps, true)
	    end

       else if opid = "TERMOF"
       then (apply_conv (TagC (mk_tag_term 1)) t, user, B.decr_steps steps, true)

       else if opid = "variable"
       then (t, user, steps, false)

       else if opid = "!abstraction"
       then let val [(v1,u1),(v2,u2),(v3,u3)] = principals
	    in (T.mk_nuprl_iabstraction_term v2 v3, user, steps, false)
	    end

       else if opid = "!library"
       then let val [([],q),([],u)] = non_principals
	    in (q, u, steps, true)
	    end

       else if is_abstraction_term lib t
       then (unfold_ab t, get_found_user user, B.decr_steps steps, true)

       else (t, user, steps, false)

    end

(* EVALUATOR1 - recursive descent *)
fun Evaluator1 lib steps term =

    let fun EvalList (terms, steps, lib) eval =
	    foldl (fn ((vars,t), (steps,terms)) =>
		      if List.null vars
		      then let val (t1,steps1,(l,u)) = eval (t,steps, lib)
			   in (steps1, terms @ [(t1,u)])
			   end
		      else raise Fail "EvalList")
		  (steps,[])
		  terms

	fun Eval (t, steps, lib) =
	    let val (opr,subterms) = T.dest_term t
		val (opid,params) = opr
		val b = E.is_eval_all opid
		val eval = if b then EvalAll o Eval else Eval
		(*val _ = print (Bool.toString b ^ "\n" ^ T.toStringTerm t ^ "\n")*)
		val (principals,non_principals) = B.split (E.num_principals opid) subterms
		val (steps1,terms) = EvalList (principals, steps, lib) eval
		fun abort () = (T.mk_nuprl_term opr ((map (fn (t,_) => ([],t)) terms) @ non_principals),
				steps1,
				lib)
	    in let val (t',u,steps',ev) = NextStepEval1 t steps terms non_principals steps1 lib
	       in (if ev then Eval else fn x => x) (t', steps', (#1 lib, u))
	       end handle _ => abort ()
	    end

	and EvalAll (t, steps, lib) =
	    let val (opr,subterms) = T.dest_term t
		val (opid,params) = opr
		val (principals,non_principals) = B.split (E.num_principal_all opid) subterms
		val (steps1,terms) = EvalList (principals, steps, lib) (EvalAll o Eval)
	    in (T.mk_nuprl_term opr ((map (fn (t,_) => ([],t)) terms) @ non_principals),
		steps1,
		lib)
	    end

	val (answer,num,_) = Eval (term, steps, lib)

    in (strip_ilib answer, steps - num)
    end

end
