(* Copyright 2011 Cornell University
 *
 *
 * This file is part of EventML - a tool aiming at specifying
 * distributed protocols in an ML like language.  It is an interface
 * to the logic of events and is compiled into Nuprl.  It is written
 * by the NUPRL group of Cornell University, Ithaca, NY.
 *
 * EventML is a free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EventML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EventML.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  o Authors:     Vincent Rahli
 *  o Affiliation: Cornell University, NUPRL group
 *  o Date:        20 May 2011
 *  o File name:   Evaluator3.sml
 *  o Description: .
 *)


structure Evaluator3 = struct

open EvaluatorsBackground

val empty_env = mk_term ("env", []) []
(* NOTE: Where is the env comming from? *)

fun init_continuation steps = fn (v, env, num) => (closure2term (v, env), steps - num)

(* EVALUATOR3 - cps (continuation passing style) transformation *)
fun Evaluator3' steps cbv_all term env k =
    let fun Eval steps term env k =
	    let val op_term = opid_of_term term
	    in if member op_term ["add", "subtract", "multiply", "divide", "remainder"]
	       then let val (opid, [t1, t2]) = dest_simple_term term
			fun F1 f steps = Eval steps t1 env (fn (v1, env1, steps1) => f v1 env1 steps1)
			fun F2 f v1 env1 steps1 = Eval steps1 t2 env (fn (v2, env2, steps2) => f v1 env1 v2 env2 steps2)
			fun F3 v1 env1 v2 env2 steps2 =
			    k (let val steps3 = decr_steps steps2
			       in (do_primitive_int_op opid v1 v2, empty_env(**), steps3)
			       end
			       handle _ => (mk_simple_term opid [closure2term (v1, env1)(**),
								 closure2term (v2, env2)(**)],
					    env(**),
					    steps2))
		    in F1 (F2 F3) steps
		    end
	       else if op_term = "minus"
	       then let val (opid, [t1]) = dest_simple_term term
			fun F1 f steps = Eval steps t1 env (fn (v1, env1, steps1) => f v1 env1 steps1)
			fun F2 v env1 steps1 =
			    k (if is_term "natural_number" v
			       then (mk_simple_term op_term [v], env1(**), steps1)
			       else let val steps2 = decr_steps steps1
				    in (do_primitive_minus v , env1(**), steps2)
				    end
				    handle _ => (mk_simple_term opid [v], env(**), steps1))
		    in F1 F2 steps
		    end
	       else if member op_term ["less", "int_eq"]
	       then let val (opid, [t1, t2, t3, t4]) = dest_simple_term term
			fun F1 f steps = Eval steps t1 env (fn (v1, env1, steps1) => f v1 env1 steps1)
			fun F2 f v1 env1 steps1 = Eval steps1 t2 env (fn (v2, env2, steps2) => f v1 env1 v2 env2 steps2)
			fun F3 v1 env1 v2 env2 steps2 =
			    let val steps3 = decr_steps steps2
			    in if do_primitive_cmp opid v1 v2
			       then Eval steps3 t3 env k
			       else Eval steps3 t4 env k
			    end handle _ => k (mk_simple_term opid [closure2term (v1, env1)(**),
								    closure2term (v2, env2)(**),
								    t3,
								    t4],
					       env(**),
					       steps2)
		    in F1 (F2 F3) steps
		    end
	       else if op_term = "atom_eq"
	       then let val (operator, [([], t1),
					([], t2),
					([], t3),
					([], t4)]) = dest_term term
			fun F1 f steps = Eval steps t1 env (fn (v1, env1, steps1) => f v1 env1 steps1)
			fun F2 f v1 env1 steps1 = Eval steps1 t2 env (fn (v2, env2, steps2) => f v1 env1 v2 env2 steps2)
			fun F3 v1 env1 v2 env2 steps2 =
			    let val steps3 = decr_steps steps2
				val n = (firstnat term) handle _ => 0
			    in if compare_atomn n v1 v2
			       then Eval steps3 t3 env k
			       else Eval steps3 t4 env k
			    end handle _ => k (mk_term operator [([], closure2term (v1, env1)(**)),
								 ([], closure2term (v2, env2)(**)),
								 ([], t3),
								 ([], t4)],
					       env(**),
					       steps2)
		    in F1 (F2 F3) steps
		    end
	       else if member op_term ["isinr", "isinl", "ispair", "isint", "islambda", "isatom2"]
	       then let val (opid, [t1, t2, t3]) = dest_simple_term term
			fun F1 f steps = Eval steps t1 env (fn (v1, env1, steps1) => f v1 env1 steps1)
			fun F2 v1 env1 steps1 =
			    let val steps2 = decr_steps steps1
			    in if do_primitive_test opid v1
			       then Eval steps2 t2 env k
			       else Eval steps2 t3 env k
			    end handle _ => k (mk_simple_term opid [closure2term (v1, env1)(**),
								    t2,
								    t3],
					       env(**),
					       steps1)
		    in F1 F2 steps
		    end
	       else if op_term = "spread"
	       then let val (operator, [([], pair), ([x, y], B)]) = dest_term term
			fun F1 f steps = Eval steps pair env (fn (v1, env1, steps1) => f v1 env1 steps1)
			fun F2 v1 env1 steps1 =
			    let val steps2 = decr_steps steps1
				val (a, b) = dest_pair v1
				val env2 = add_env_binding (add_env_binding env x a env1) y b env1(**)
			    in Eval steps2 B env2 k
			    end handle _ => k (mk_term operator [([], closure2term (v1, env1)(**)),
								 ([x, y], B)],
					       env(**),
					       steps1)
		    in F1 F2 steps
		    end
	       else if op_term = "decide"
	       then let val (operator, [([], d), ([x], A), ([y], B)]) = dest_term term
			fun F1 f steps = Eval steps d env (fn (v1, env1, steps1) => f v1 env1 steps1)
			fun F2 v env1 steps1 =
			    let val steps2 = decr_steps steps1
			    in if is_term "inl" v
			       then let val env2 = add_env_binding env x (subtermn 1 v) env1(**)
				    in Eval steps2 A env2 k
				    end
			       else if is_term "inr" v
			       then let val env2 = add_env_binding env y (subtermn 1 v) env1(**)
				    in Eval steps2 B env2 k
				    end
			       else raise FAIL
			    end handle _ => k (mk_term operator [([], closure2term (v, env1))(**),
								 ([x], A),
								 ([y], B)],
					       env,
					       steps1)
		    in F1 F2 steps
		    end
	       else if op_term = "apply"
	       then if steps < 0 andalso is_term "ycomb" (subtermn 1 term)
		    (* NOTE: Why do that only in negative-steps mode? *)
		    then let val steps1 = decr_steps steps
			     val f = subtermn 2 term
			 in Eval steps1 (mk_apply_term f term) env k
			 end handle _ => k (term, env(**), steps)
		    else let val (operator, [([], function), ([], arg)]) = dest_term term
			     fun F1 f steps = Eval steps function env (fn (v1, env1, steps1) => f v1 env1 steps1)
			     fun F2 f fenv steps1 =
				 let val steps2 = decr_steps steps1
				     val (x, B) = dest_lambda f
				     val env' = add_env_binding fenv x arg env
				 in Eval steps2 B env' k
				 end handle _ => k (mk_apply_term (closure2term (f, fenv)) arg, env, steps1)
			 in F1 F2 steps
			 end
	       else if op_term = "callbyvalue"
	       then let val (operator, [([], arg), ([x], B)]) = dest_term term
			fun F1 f steps = Eval steps arg env (fn (v1, env1, steps1) => f v1 env1 steps1)
			fun F2 v env1 steps1 =
			    let val steps2 = decr_steps steps1
			    in if is_primitive_value v
			       then let val env' = add_env_binding env x v env1
				    in Eval steps2 B env' k
				    end
			       else raise FAIL
			    end handle _ => k (mk_term operator [([], closure2term (v, env1)),
								 ([x], B)],
					       env,
					       steps1)
		    in F1 F2 steps
		    end
	       else if op_term = "callbyvalueall"
	       then let val (operator, [([], arg), ([x], B)]) = dest_term term
			fun F1 f steps = Evaluator3' steps true arg env (fn (v1, env1, steps1) => f v1 env1 steps1)
			fun F2 v1 env1 steps1 =
			    let val steps2 = decr_steps steps1
			    in if is_complete_primitive_value v1
			       then let val env' = add_env_binding env x v1 env1
				    in Evaluator3' steps2 cbv_all B env' k
				    end
			       else raise FAIL
			    end handle _ => k (mk_term operator [([], closure2term (v1, env1)),
								 ([x], B)],
					       env(**),
					       steps1)
		    in F1 F2 steps
		    end
	       else if op_term = "list_ind"
	       then let val (operator, [([], L),
					([], nilcase),
					([x, xs, r], conscase)]) = dest_term term
			fun F1 f steps = Eval steps L env (fn (v1, env1, steps1) => f v1 env1 steps1)
			fun F2 v env1 steps1 =
			    let val steps2 = decr_steps steps1
			    in if is_term "nil" v
			       then Eval steps2 nilcase env k
			       else if is_term "cons" v
			       then let val va = subtermn 1 v
					val vb = subtermn 2 v
					val env2 = add_env_binding env xs vb env1
					val env3 = add_env_binding env2 x va env1
					val w    = fresh_variable term
					val env4 = add_env_binding env w vb env1
					val t2 = mk_term operator [([],         mk_variable_term w),
								   ([],         nilcase),
								   ([x, xs, r], conscase)]
					val env4 = add_env_binding env3 r t2 env4
				    in Eval steps2 conscase env4 k
				    end
			       else raise FAIL
			    end handle _ => k (mk_term operator [([],         closure2term (v, env1)(**)),
								 ([],         nilcase),
								 ([x, xs, r], conscase)],
					       env(**),
					       steps1)
		    in F1 F2 steps
		    end
	       else if op_term = "ind"
	       then let val (operator, [([], i),
					([x, rd], downcase),
					([], basecase),
					([y, ru], upcase)]) = dest_term term
			fun F1 f steps = Eval steps i env (fn (v1, env1, steps1) => f v1 env1 steps1)
			fun F2 v qenv steps1 =
			    let val steps2 = decr_steps steps1
				val n = dest_integer v
			    in if n = 0
			       then Eval steps2 basecase env k
			       else if n > 0
			       then let val t2 = mk_term operator [([],      mk_integer_term (n-1)),
								   ([x, rd], downcase),
								   ([],      basecase),
								   ([y, ru], upcase)]
					val env1 = add_env_binding env y v qenv(**)
					val env2 = add_env_binding env1 ru t2 env(**)
				    in Eval steps2 upcase env2 k
				    end
			       else let val t2 = mk_term operator [([],      mk_integer_term (n+1)),
								   ([x, rd], downcase),
								   ([],      basecase),
								   ([y, ru], upcase)]
					val env1 = add_env_binding env x v qenv(**)
					val env2 = add_env_binding env1 rd t2 env(**)
				    in Eval steps2 downcase env2 k
				    end
			    end handle _ => k (mk_term operator [([],      closure2term (v, qenv)(**)),
								 ([x, rd], downcase),
								 ([],      basecase),
								 ([y, ru], upcase)],
					       env(**),
					       steps1)
		    in F1 F2 steps
		    end
	       else if op_term = "rec_ind"
	       then let val (operator, [([], arg), ([f, x], B)]) = dest_term term
			val function = mk_lambda_term x (mk_term operator [([], mk_variable_term x), ([f, x], B)])
		    in let val steps1 = decr_steps steps
			   val env1 = add_env_binding env x arg env(**)
			   val env2 = add_env_binding env1 f function env(**)
		       in Eval steps1 B env2 k
		       end handle _ => k (term, env, steps)
		    end
	       else if op_term = "limited_type_case"
	       then let fun F1 f steps = Eval steps (subtermn 1 term) env (fn (v1, env1, steps1) => f v1 env1 steps1)
			fun F2 v ve steps1 =
			    let val steps2 = decr_steps steps1
				val (t2, env2(**)) = limited_type_case2(**) term v ve env(**)
			    in Eval steps2 t2 env2 k
			    end handle _ => k (limited_type_case_no_redex2(**) term v ve(**), env(**), steps1)
		    in F1 F2 steps
		    end
	       else if op_term = "TERMOF"
	       then let val steps1 = decr_steps steps
		    in Eval steps1 (apply_conv (TagC (mk_tag_term 1)) term) env k
		    end handle _ => k (term, env(**), steps)
	       else if false (*is_abstraction_term term*)
	       then let val steps1 = decr_steps steps
		    in Eval steps1 (unfold_ab term) env k
		    end handle _ => k (term, env(**), steps)
	       else if op_term = "variable"
	       then let val (a, e) = lookup_binding env (dest_variable term)(**)
		    in Eval steps a e k
		    end handle _ => k (term, env(**), steps)
	       else if member op_term ["pair", "cons"]
	       then if cbv_all
		    then let val (opid, [t1, t2]) = dest_simple_term term
			     fun F1 f steps = Eval steps t1  env (fn (v1, env1, steps1) => f v1 env1 steps1)
			     fun F2 f v1 env1 steps1 = Eval steps1 t2 env1 (fn (v2, env2, steps2) => f v1 env1 v2 env2 steps2)
			     fun F3 v1 env1 v2 env2 steps2 =
				 k (mk_simple_term opid [v1, v2], env2, steps2)
			 in F1 (F2 F3) steps
			 end
		    else k (term, env, steps)
	       else if member op_term ["inl", "inr"]
	       then if cbv_all
		    then let val (opid, [t1]) = dest_simple_term term
			     fun F1 f steps = Eval steps t1 env (fn (v1, env1, steps1) => f v1 env1 steps1)
			     fun F2 v1 env1 steps1 =
				 k (mk_simple_term opid [v1], env1, steps1)
			 in F1 F2 steps
			 end
		    else k (term, env, steps)
	       else k (term, env, steps)
	    (* NOTE: there is another ite concerning the bound variables of term. *)
	    end
    in Eval steps term env k
    end

fun Evaluator3 steps term =
    Evaluator3' steps false term empty_env (init_continuation steps)

end
