(* Copyright 2011 Cornell University
 *
 *
 * This file is part of EventML - a tool aiming at specifying
 * distributed protocols in an ML like language.  It is an interface
 * to the logic of events and is compiled into Nuprl.  It is written
 * by the NUPRL group of Cornell University, Ithaca, NY.
 *
 * EventML is a free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EventML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EventML.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  o Authors:     Vincent Rahli
 *  o Affiliation: Cornell University, NUPRL group
 *  o Date:        20 May 2011
 *  o File name:   ToNuprl.sml
 *  o Description: Compiles EventML to Nuprl.
 *)


structure ToNuprl :> TONUPRL = struct

structure A  = Ast
structure E  = Env
structure D  = Deps
structure EN = Enum
structure T  = ListFormat
structure EH = LibBase
structure NT = NuprlTerms


val toStringTerm = NT.toStringTerm

(* ------ IDENTIFIERS ------ *)

structure IDS = A.IDS

fun get_all_ids_bterm (NT.B_TERM (variables, term)) =
    IDS.addList (get_all_ids term, variables)

and get_all_ids (term as NT.TERM (((opid, tag), params), bterms)) =
    let val set =
	    if NT.is_nuprl_variable_term term
	    then IDS.add (IDS.singleton opid, NT.dest_variable term)
	    else IDS.singleton opid
    in foldr (fn (bterm, set) => IDS.union (set, get_all_ids_bterm bterm))
	     set
	     bterms
    end

fun get_all_ids_list lst =
    let val some_ids =
	    ["bag-map", "bag-eq", "lambda", "pair", "eclass",
	     "normal-locally-programmable", "programmable",
	     "variable"]
    in foldr (fn (term, set) => IDS.union (set, get_all_ids term))
	     (IDS.addList (IDS.empty, some_ids))
	     lst
    end


(* ------ CONSTRUCTORS ------ *)

fun mk_nuprl_eta1_term v term =
    let val set = get_all_ids_list [term]
	val id  = A.newIdIdSet set v
	val nid = NT.mk_nuprl_variable_term id
	val app = NT.mk_nuprl_apply_term term nid
    in NT.mk_nuprl_lambda_term id term
    end

fun mk_nuprl_eta2_term v1 v2 term =
    let val set  = get_all_ids_list [term]
	val id1  = A.newIdIdSet set v1
	val id2  = A.newIdIdSet set v2
	val nid1 = NT.mk_nuprl_variable_term id1
	val nid2 = NT.mk_nuprl_variable_term id2
	val app1 = NT.mk_nuprl_apply_term term nid1
	val app2 = NT.mk_nuprl_apply_term app1 nid2
	val lam1 = NT.mk_nuprl_lambda_term id2 app2
	val lam2 = NT.mk_nuprl_lambda_term id1 lam1
    in lam2
    end

fun mk_nuprl_bag_deq_term deq =
    let val set  = get_all_ids_list [deq]
	val id1  = A.newIdIdSet set "x"
	val id2  = A.newIdIdSet set "y"
	val nid1 = NT.mk_nuprl_variable_term id1
	val nid2 = NT.mk_nuprl_variable_term id2
	val bdeq = NT.mk_nuprl_simple_term "bag-eq"  [deq, nid1, nid2]
	val lam1 = NT.mk_nuprl_lambda_term id2 bdeq
	val lam2 = NT.mk_nuprl_lambda_term id1 lam1
    in lam2
    end


(* ------ EXPORT E# TO NUPRL ------ *)

val minStr = "this term should have been transformed"

(* We have this environment as a reference to avoid passing it around in ToNuprl. *)
val unifenv = ref (EN.initSolved ())
fun setUnifEnv x  = unifenv := x
fun getUnifEnv () = !unifenv

(* Bound variables *)
structure BOUNDS = BinarySetFn(type ord_key = string val compare = String.compare)

val bounds = ref BOUNDS.empty
fun isBound v = BOUNDS.member (!bounds, v)
fun addBound v =
    if isBound v
    then false
    else (bounds := BOUNDS.add (!bounds, v); true)
fun rmBound b v =
    if b
    then (bounds := BOUNDS.delete (!bounds, v))
	 handle LibBase.NotFound => raise EH.Impossible "rmBound"
    else ()

(* This map is used to record how many times an identifier has been declared in
 * a piece of code.  If we're converting a declaration and we know that id has
 * been declared twice then we know that to get its binding stored in the
 * unifenv, we need to look at the third binding of id from the start. *)
structure NBID = BinaryMapFn(type ord_key = string val compare = String.compare)

val nbId = ref (NBID.empty : int NBID.map)
fun addOneId id =
    case NBID.find (!nbId, id) of
	NONE   => nbId := NBID.insert (!nbId, id, 1)
      | SOME n => nbId := NBID.insert (!nbId, id, n + 1)
fun getNbOfId id =
    case NBID.find (!nbId, id) of
	NONE => 0
      | SOME n => n
fun resetNbId () = nbId := NBID.empty

(* This map is used to record the arity of imported constants. *)
structure ARITY = BinaryMapFn(type ord_key = string val compare = String.compare)

val arity = ref (ARITY.empty : (string * (string * string list) list) ARITY.map)
fun addArity id name ar = arity := ARITY.insert (!arity, id, (name, ar))
fun getArity id =
    if isBound id
    then NONE
    else ARITY.find (!arity, id)
fun resetArity () = arity := ARITY.empty

(* This map is used to record the parameters used in the declarared function.
 * The number of parameters used in a function is then used to define the arity
 * of the function. *)
structure CONS = BinaryMapFn(type ord_key = string val compare = String.compare)

(* userFP stands for 'user functions with their parameters' *)
val userFP = ref (CONS.empty : string list CONS.map)
fun addUserFP id params = userFP := CONS.insert (!userFP, id, params)
fun getUserFP id =
    if isBound id
    then NONE
    else CONS.find (!userFP, id)
fun resetUserFP () = userFP := CONS.empty

(* This set is used to record the parameters added to the declarations' bodies *)
structure DECFP = BinarySetFn(type ord_key = string val compare = String.compare)

val decFP = ref DECFP.empty
fun addDecFP param = decFP := DECFP.add (!decFP, param)
fun addDecFPList params = decFP := DECFP.addList (!decFP, params)
fun getDecFP () = DECFP.listItems (!decFP)
fun resetDecFP () = decFP := DECFP.empty

(*
(* This set is used to record the type parameters used in declarations' types. *)
structure PTYPES = BinarySetFn(type ord_key = string val compare = String.compare)

val ptypes = ref PTYPES.empty
fun addPType tparam = ptypes := PTYPES.add (!ptypes, tparam)
fun getPTypes () = PTYPES.listItems (!ptypes)
fun resetPTypes () = ptypes := PTYPES.empty
*)

(* This map is used to record the defined parameters along with their types *)
structure PARAMS = BinaryMapFn(type ord_key = string val compare = String.compare)

val params = ref (PARAMS.empty : E.scheme PARAMS.map)
fun addParam param typ = params := PARAMS.insert (!params, param, typ)
fun getParam param = PARAMS.find (!params, param)
fun isParam  param = Option.isSome (PARAMS.find (!params, param))
fun resetParams () = params := PARAMS.empty

(*
(* This set is used to record type parameters. *)
structure TPARAMS = BinarySetFn(type ord_key = string val compare = String.compare)

val tparams = ref TPARAMS.empty
fun addTParam tparam = tparams := TPARAMS.add (!tparams, tparam)
fun isTParam  tparam = TPARAMS.member (!tparams, tparam)
fun resetTParams  () = tparams := TPARAMS.empty
*)

(* This map is used to record the equality deciders *)
structure EQDEC = BinaryMapFn(type ord_key = string val compare = String.compare)

val eqdec = ref (EQDEC.empty : string EQDEC.map)
fun addEqDec tc eq = eqdec := EQDEC.insert (!eqdec, tc, eq)
fun rmEqDec tc = (eqdec := #1 (EQDEC.remove (!eqdec, tc))) handle LibBase.NotFound => ()
fun getEqDec tc = EQDEC.find (!eqdec, tc)
fun resetEqDec () = eqdec := EQDEC.empty

(* This list is used to record all the headers as nuprl terms.
 * We do that to generate the full list of headers of an E# specification. *)
val headers = ref ([] : (string * NT.nuprl_term) list)
fun addHeader p = headers := p :: !headers
fun getHeaders () = !headers
fun resetHeaders () = headers := []

(* Object id of the esharp file within Nuprl. *)
val obidref = ref ""
fun setObid obid = obidref := obid
fun getObid () = !obidref
fun resetObid () = obidref := ""

(* Prefix to use for the generated Nuprl objects. *)
val prefref = ref ""
fun setPrefix pref = prefref := pref
fun getPrefix () = !prefref
fun resetPrefix () = prefref := ""

fun get_user_name opid = getPrefix () ^ opid

fun mk_nuprl_user_term_lvlop opid terms lvlop =
    let val obid    = getObid ()
	val name    = get_user_name opid
	val bterms  = (map (fn t => ([], t)) terms)
	val params1 = if obid = "" then [] else [NT.mk_nuprl_obid_parameter obid]
	val params2 = case lvlop of SOME i => [NT.mk_nuprl_level_exp_parameter i] | NONE => []
    in  NT.mk_nuprl_term (name, params1 @ params2) bterms
    end

fun mk_nuprl_user_term opid terms = mk_nuprl_user_term_lvlop opid terms NONE

fun reset () =
    (resetNbId    ();
     resetArity   ();
     resetParams  ();
     (*resetTParams ();*)
     resetEqDec   ();
     resetUserFP  ();
     resetDecFP   ();
     (*resetPTypes  ();*)
     resetHeaders ();
     resetObid    ();
     resetPrefix  ();
     ())

fun getArityArgs (A.N {kind = (A.ARGS, A.ARGS_EM),   label, value, regions, parents, children = []}) = []
  | getArityArgs (A.N {kind = (A.ARGS, A.ARGS_PSEQ), label, value, regions, parents, children})      =
    map (fn (A.N {kind = (A.ARG, A.ARG_A), label, value, regions, parents, children = [id, args, ty]}) =>
	    (case args of
		 A.N {kind = (A.ARGS, A.ARGS_EM), label, value, regions, parents, children = []} => (A.getIdIdent id, [])
	       | A.N {kind = (A.ARGS, A.ARGS_LSEQ), label, value, regions, parents, children}    =>
		 let val lst = map (fn (A.N {kind = (A.ARG, A.ARG_A), label, value, regions, parents, children = [id, args, ty]}) => A.getIdIdent id
				     | _ => raise EH.Impossible "unexpected term")
				   children
		 in (A.getIdIdent id, lst)
		 end
	       | _ => raise EH.Impossible "unexpected term")
	  | _ => raise EH.Impossible "unexpected term")
	children
  | getArityArgs _ = raise EH.Impossible "unexpected term"

fun getArityTyvarSeq (A.N {kind = (A.TYPEVARSEQ, A.TYPEVARSEQ_EM),  label, value, regions, parents, children}) = 0
  | getArityTyvarSeq (A.N {kind = (A.TYPEVARSEQ, A.TYPEVARSEQ_ONE), label, value, regions, parents, children}) = 1
  | getArityTyvarSeq (A.N {kind = (A.TYPEVARSEQ, A.TYPEVARSEQ_SEQ), label, value, regions, parents, children}) = List.length children
  | getArityTyvarSeq term = raise EH.Impossible "unexpected term"

fun flattenExpApp (A.N {kind = (A.EXP, A.EXP_APP), label, value, regions, parents, children = [exp, atexp]}) =
    let val (term, terms) = flattenExpApp exp
    in (term, terms @ [atexp])
    end
  | flattenExpApp term = (term, [])

fun toNuprlListAtoms [] = NT.mk_nuprl_nil_term
  | toNuprlListAtoms (x :: xs) =
    let val t = NT.mk_nuprl_token_term x
    in NT.mk_nuprl_cons_term t (toNuprlListAtoms xs)
    end

fun toAtomList atoms =
    case A.getKind atoms of
	(A.ATOMS, A.ATOMS_ATOMS) => toNuprlListAtoms (A.splitAtoms (A.getValue atoms))
      | _ => raise EH.Impossible ""

fun getIdPat n pat =
    case A.getIdPat pat of
	SOME id => A.getIdIdent id
      | NONE => (print (Int.toString n ^ "\n" ^ A.toString pat); raise EH.Impossible "")

(* TODO: we don't need this function, we can just gather the free params, while
 * converting the E# code, because we anyway make sure that the bound variables
 * don't clash with the parameter names. *)
fun getFreeParams binders (A.N {kind = (A.ID, A.ID_VID), label, value, regions, parents, children = []})=
    if not (isParam value)
       orelse
       List.exists (fn x => x = value) binders
    then []
    else [value]
  | getFreeParams binders (A.N {kind = (A.EXP, A.EXP_LET), label, value, regions, parents, children = [bind, exp]}) =
    (case bind of
	 A.N {kind = (A.BIND, A.BIND_DEC), label, value, regions, parents, children = [f, a, e]} =>
	 if A.termIsEmParam a
	 then let val fid = A.getIdIdent f
		  val params1 = getFreeParams binders e
		  val params2 = getFreeParams (fid :: binders) exp
	      in params1 @ params2
	      end
	 else raise EH.Impossible minStr
       | A.N {kind = (A.BIND, A.BIND_PAT), label, value, regions, parents, children = [p, e]} =>
	 let val idents =
		 case p of
		     A.N {kind = (A.PAT, A.PAT_ATPAT), label, value, regions, parents, children = [atpat]} =>
		     (case atpat of
			  A.N {kind = (A.ATPAT, A.ATPAT_TUPLE), label, value, regions, parents, children} => map (getIdPat 1) children
			| A.N {kind = (A.ATPAT, A.ATPAT_ID), label, value, regions, parents, children = [id]} => [A.getIdIdent id]
			| term => (print (A.toString term ^ "\n"); raise EH.Impossible minStr))
		   | _ => raise EH.Impossible minStr
	     val params1 = getFreeParams binders e
	     val params2 = getFreeParams (idents @ binders) exp
	 in params1 @ params2
	 end
       | _ => raise EH.Impossible minStr)
  | getFreeParams binders (A.N {kind = (A.EXP, A.EXP_LETR), label, value, regions, parents, children = [bind, exp]}) =
    (case bind of
	 A.N {kind = (A.BIND, A.BIND_DEC), label, value, regions, parents, children = [f, a, e]} =>
	 if A.termIsEmParam a
	 then let val fid = A.getIdIdent f
		  val params1 = getFreeParams (fid :: binders) exp
		  val params2 = getFreeParams (fid :: binders) e
	      in params1 @ params2
	      end
	 else raise EH.Impossible minStr
       | _ => raise EH.Impossible minStr)
  | getFreeParams binders (A.N {kind = (A.EXP, A.EXP_LAMBDA), label, value, regions, parents, children = [pat, exp]}) =
    getFreeParams ((getIdPat 2 pat) :: binders) exp
  | getFreeParams binders (A.N {kind = (A.EXP, A.EXP_BINDING), label, value, regions, parents, children = [exp1, pat, exp2]}) =
    let val params1 = getFreeParams binders exp1
	val params2 = getFreeParams ((getIdPat 3 pat) :: binders) exp2
    in params1 @ params2
    end
  | getFreeParams binders (A.N {kind, label, value, regions, parents, children}) = List.concat (map (getFreeParams binders) children)

structure BTV = BinaryMapFn(type ord_key = E.ityvar val compare = Int.compare)

fun toNuprlIty i (E.ITYVAR (ityvar, eq)) ityvarset etyvarset =
    (case BTV.find (ityvarset, ityvar) of
	 SOME str => NT.mk_nuprl_variable_term str
       | NONE => raise EH.Impossible "type variables are not allowed to be converted to nuprl")
  (* We can have a type variable because we have a genuine generality or the
   * ityvar can come from a mono bind.  We might have to do more that recording
   * the type variable associated to a label.
   * We should be able to handle:
   *
   *     MSGS (``foo` : 'a List) ;;
   *     let m = MSG(``foo``, []) ;;
   *
   * But we don't. *)
  | toNuprlIty i (E.ITYETV (etyvar, lab)) ityvarset etyvarset =
    if E.isinETVS (etyvarset, etyvar)
    then NT.mk_nuprl_variable_term etyvar
    else raise EH.Impossible "type variables are not allowed to be converted to nuprl "
  | toNuprlIty i (E.ITYCON (E.ITYSEQSEQ (itys, _), E.ITYCONNAM (tyconnam, _, _))) ityvarset etyvarset =
    let val n_itys = map (fn ity => toNuprlIty i ity ityvarset etyvarset) itys
    in case (tyconnam, n_itys) of
           ("*",     nt1 :: nt2 :: nts) => foldl (fn (nty, prd) => NT.mk_nuprl_prod_term prd nty) (NT.mk_nuprl_prod_term nt1 nt2) nts
	 | ("*",     []) => NT.mk_nuprl_unit_term
	 | ("->",    [nt1, nt2]) => NT.mk_nuprl_fun_term   nt1 nt2
	 | ("+",     [nt1, nt2]) => NT.mk_nuprl_union_term nt1 nt2
	 | ("List",  [nt]) => NT.mk_nuprl_list_term  nt
	 | ("Class", [nt]) => NT.mk_nuprl_class_term nt
	 | ("Bag",   [nt]) => NT.mk_nuprl_bag_term   nt
	 | ("Deq",   [nt]) => NT.mk_nuprl_deq_term   nt
	 | ("Int",   []) => NT.mk_nuprl_int_term
	 | ("Bool",  []) => NT.mk_nuprl_bool_term
	 | ("Real",  []) => NT.mk_nuprl_real_term
	 | ("Msg",   []) => NT.mk_nuprl_msg_term "i"
	 | ("Unit",  []) => NT.mk_nuprl_unit_term
	 | ("Atom",  []) => NT.mk_nuprl_atom_term
	 | ("Type",  []) => NT.mk_nuprl_type_term "i"
	 | ("Prop",  []) => NT.mk_nuprl_prop_term "i"
	 | ("Nat",   []) => NT.mk_nuprl_nat_term
	 | ("Loc",   []) => NT.mk_nuprl_loc_term
	 | (tycon, args) =>
	   if isParam tycon
	   then let val _ = addDecFP tycon
		in foldl (fn (arg, term) => NT.mk_nuprl_apply_term term arg)
			 (NT.mk_nuprl_variable_term tycon)
			 args
		end
	   else NT.mk_nuprl_simple_term tycon args
    end
  | toNuprlIty i (E.ITYDEP (ity, _)) ityvarset etyvarset = toNuprlIty i ity ityvarset etyvarset
  | toNuprlIty i ity _ _ = (print ("*** (" ^ Int.toString i ^ ") " ^ E.toStringIty ity ^ "\n"); raise EH.Impossible "the type should be fully built up, no variable")

(* transforms a type into an equality decider *)
fun toDeq (E.ITYVAR (ityvar, eq)) =
    raise EH.Impossible "type variables are not allowed to be converted to nuprl"
  | toDeq (E.ITYETV (etyvar, lab)) =
    raise EH.Impossible "type variables are not allowed to be converted to nuprl "
  | toDeq (E.ITYCON (E.ITYSEQSEQ (itys, _), E.ITYCONNAM (tyconnam, _, _))) =
    let fun toIty ity = toNuprlIty 0 (E.stripIty ity) BTV.empty E.emptyETVS
	val n_deqs = map (fn ity => (toIty ity, toDeq ity)) itys
    in case (tyconnam, n_deqs) of
	   ("*", []) => NT.mk_nuprl_unit_deq_term
	 | ("*",     x :: y :: deqs) =>
	   (case List.rev n_deqs of
		(nty1, deq1) :: (nty2, deq2) :: deqs =>
		let val nty0 = NT.mk_nuprl_prod_term nty2 nty1
		    val deq0 = NT.mk_nuprl_product_deq_term nty2 nty1 deq2 deq1
		    val (_, deq) =
			foldl (fn ((nty', deq'), (nty, deq)) =>
				  (NT.mk_nuprl_prod_term nty' nty,
				   NT.mk_nuprl_proddeq_term deq' deq))
			      (nty0, deq0)
			      deqs
		in deq
		end
	      | _ => raise EH.Impossible "product type should have at least 2 arguments")
	 | ("->",    [(nty1, deq1), (nty2, deq2)]) => raise EH.Impossible "functional type does not have an equality decider"
	 | ("+",     [(nty1, deq1), (nty2, deq2)]) => NT.mk_nuprl_union_deq_term nty1 nty2 deq1 deq2
	 | ("Class", [(nty, deq)]) => raise EH.Impossible "Class does not have an equality decider"
	 | ("List",  [(nty, deq)]) => NT.mk_nuprl_list_deq_term deq
	 | ("Bag",   [(nty, deq)]) => mk_nuprl_bag_deq_term  deq
	 | ("Int",   []) => NT.mk_nuprl_int_deq_term
	 | ("Bool",  []) => NT.mk_nuprl_bool_deq_term
	 | ("Real",  []) => raise EH.Impossible "Real does not have an equality decider"
	 | ("Msg",   []) => raise EH.Impossible "Msg does not have an equality decider"
	 | ("Unit",  []) => NT.mk_nuprl_unit_deq_term
	 | ("Atom",  []) => NT.mk_nuprl_atom_deq_term
	 | ("Type",  []) => raise EH.Impossible "Type does not have an equality decider"
	 | ("Nat",   []) => NT.mk_nuprl_nat_deq_term
	 | ("Loc",   []) => NT.mk_nuprl_loc_deq_term
	 | (_, _) =>
	   (case getEqDec tyconnam of
		NONE =>
		let val str = "no equality decider for " ^ tyconnam
		    val _   = print (str ^ "\n")
		in raise EH.Impossible str
		end
	      | SOME deq =>
		if isParam deq
		then let val _ = addDecFPList [deq]
		     in NT.mk_nuprl_applies_term (NT.mk_nuprl_variable_term deq) (map (fn (_, x) => x) n_deqs)
		     end
		else NT.mk_nuprl_applies_term (NT.mk_nuprl_simple_term deq [])  (map (fn (_, x) => x) n_deqs))
    end
  | toDeq ity = (print ("*** " ^ E.toStringIty ity ^ "\n"); raise EH.Impossible "the type should be fully built up, no variable")

(*
(* transforms a type and two terms into a boolean. *)
fun toEqDeq nt1 nt2 (E.ITYVAR (ityvar, eq)) =
    raise EH.Impossible "type variables are not allowed to be converted to nuprl"
  | toEqDeq nt1 nt2 (E.ITYETV (etyvar, lab)) =
    raise EH.Impossible "type variables are not allowed to be converted to nuprl "
  | toEqDeq nt1 nt2 (E.ITYCON (E.ITYSEQSEQ (itys, _), E.ITYCONNAM (tyconnam, _, _))) =
    let val n_deqs = map (fn ity => toDeq ity) itys
    in case (tyconnam, n_deqs) of
           ("*",     [deq1, deq2]) => NT.mk_nuprl_eq_prod_term deq1 deq2 nt1 nt2
	 | ("->",    [deq1, deq2]) => raise EH.Impossible "-> does not have an equality decider"
	 | ("+",     [deq1, deq2]) => NT.mk_nuprl_eq_union_term deq1 deq2 nt1 nt2
	 | ("Class", [deq]) => raise EH.Impossible "Class does not have an equality decider"
	 | ("List",  [deq]) => NT.mk_nuprl_eq_list_term deq nt1 nt2
	 | ("Bag",   [deq]) => NT.mk_nuprl_eq_bag_term deq nt1 nt2
	 | ("Int",   []) => NT.mk_nuprl_eq_int_term nt1 nt2
	 | ("Bool",  []) => NT.mk_nuprl_eq_bool_term nt1 nt2
	 | ("Real",  []) => raise EH.Impossible "Real does not have an equality decider"
	 | ("Msg",   []) => raise EH.Impossible "Msg does not have an equality decider"
	 | ("Unit",  []) => raise EH.Unimplemented ""
	 | ("Atom",  []) => NT.mk_nuprl_eq_atom_term nt1 nt2
	 | ("Type",  []) => raise EH.Impossible "Type does not have an equality decider"
	 | ("Nat",   []) => NT.mk_nuprl_eq_int_term nt1 nt2
	 | ("Loc",   []) => NT.mk_nuprl_eq_loc_term nt1 nt2
	 | (_, _) =>
	   (case getEqDec tyconnam of
		NONE => raise EH.Impossible ("no equality decider for " ^ tyconnam)
	      | SOME deq =>
		let val ndeq = NT.mk_nuprl_simple_term deq n_deqs
		    val app1 = NT.mk_nuprl_apply_term ndeq nt1
		    val app2 = NT.mk_nuprl_apply_term app1 nt2
		in app2
		end)
    end
  | toEqDeq nt1 nt2 ity = (print ("*** " ^ E.toStringIty ity ^ "\n"); raise EH.Impossible "the type should be fully built up, no variable")
*)

datatype alpha = NEXT of string * (unit -> alpha)

val lstAlpha =
    ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M",
     "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z"]

fun getNewTyVarName [] suff () = getNewTyVarName lstAlpha (suff ^ "'") ()
  | getNewTyVarName (x :: xs) suff () =
    let val f = getNewTyVarName xs suff
    in NEXT (x ^ suff, f)
    end

val getNewTyVarName = getNewTyVarName lstAlpha ""

(* NOTE: idorset should be empty, we shouldn't have any or types in the types
 * of the top declarations. *)
fun toNuprlScheme (E.VPOLY ((ityvarset, etyvarset, idorset), (overseq, ity))) =
    let val ity'  = E.stripIty ity
	val (nts1, ren) =
	    let val (nts, ren, _) =
		    E.foldrITVS (fn (tv, (nts, ren, f)) =>
				    let val NEXT (str, f') = f ()
				    in (str :: nts, BTV.insert (ren, tv, str), f')
				    end)
				([], BTV.empty, getNewTyVarName)
				ityvarset
	    in (nts, ren)
	    end
	val nts2  = E.listItemsETVS etyvarset
	val n_ity = toNuprlIty 1 ity' ren etyvarset
    in (n_ity, nts2 @ nts1)
    end
  | toNuprlScheme (E.VODEC _) = raise EH.Impossible "overloaded schemes are not allowed to be converted into NuPrl"
  | toNuprlScheme (E.VMONO _) = raise EH.Impossible "monomorphic schemes are not allowed to be converted into NuPrl"

(* NOTE: ityvarset and idorset should be empty. *)
fun schemeParamToHead (E.VPOLY ((ityvarset, etyvarset, idorset), (overseq, ity))) =
    let val ity'  = E.stripIty ity
	val nts   = E.listItemsETVS etyvarset
	val n_ity = toNuprlIty 2 ity' BTV.empty etyvarset
    in (n_ity, nts)
    end
  | schemeParamToHead (E.VODEC _) = raise EH.Impossible "overloaded schemes are not allowed to be converted into NuPrl"
  | schemeParamToHead (E.VMONO _) = raise EH.Impossible "monomorphic schemes are not allowed to be converted into NuPrl"

structure SETVARS = BinarySetFn(type ord_key = string val compare = String.compare)

fun getVariablesNuprlType term =
    if NT.is_nuprl_variable_term term
    then SETVARS.singleton (NT.dest_variable term)
    else let val (opid, sterms) = NT.dest_simple_null_term term
	 in foldr (fn (term, vars) =>
		      SETVARS.union (getVariablesNuprlType term, vars))
		  SETVARS.empty
		  sterms
	 end

fun findNlpsInType term typ (args, set) =
    let val opid = NT.opid_of_term typ
    in if opid = "eclass"
       then let val (_, _, _, Y) = NT.dest_eclass typ
		val nlp = NT.mk_nuprl_nlp_term Y term
		val prg = NT.mk_nuprl_programmable_term Y term
	    in ([(nlp, prg)], args, set)
	    end
       else if opid = "product"
       then let val (left, right) = NT.dest_product typ
		val (nlps1, args1, set1) = findNlpsInType (NT.mk_nuprl_pi1_term term)
							  left
							  ([], set)
		val (nlps2, args2, set2) = findNlpsInType (NT.mk_nuprl_pi2_term term)
							  right
							  ([], set1)
		val nlps  = nlps1 @ nlps2
		val args' = if null nlps then [] else args
	    in (nlps, args' @ args1 @ args2, set2)
	    end
       else if opid = "function"
       then let val (left, right) = NT.dest_function typ
		val id    = A.newId ()
		val v     = NT.mk_nuprl_variable_term id
		val args' = (id, left) :: args
		val set'  = IDS.add (set, id)
	    in findNlpsInType (NT.mk_nuprl_apply_term term v) right (args', set')
	    end
       else ([], [], set)
    end

fun genNormLocProg set class nity args =
    let val opid = NT.opid_of_term nity
    in if opid = "eclass"
       then let val (msg, es, e, X) = NT.dest_eclass nity
		val v_args   = map (fn (v, typ) => NT.mk_nuprl_variable_term v) args
		val applies  = NT.mk_nuprl_applies_term class v_args
		val setvars  = getVariablesNuprlType X
		val vars     = SETVARS.listItems setvars
		(*val valueall = NT.mk_nuprl_valueall_type_term X*)
		val nlp      = NT.mk_nuprl_nlp_term X applies
		val prg      = NT.mk_nuprl_programmable_term X applies
		(*val vallimp  = NT.mk_nuprl_implies_term valueall nlp*)
		val (nlpargs, oargs, _)  =
		    foldr (fn ((v, typ), (nlps, args, set)) =>
			      let val term = NT.mk_nuprl_variable_term v
				  val (nlps', args', set') = findNlpsInType term typ ([], set)
			      in (nlps' @ nlps, args' @ args, set')
			      end)
			  ([], [], set)
			  args
		val (nlpsimp, prgsimp)  =
		    foldr (fn ((leftN, leftP), (rightN, rightP)) =>
			      (NT.mk_nuprl_implies_term leftN rightN,
			       NT.mk_nuprl_implies_term leftP rightP))
			  (nlp, prg) (*vallimp*)
			  nlpargs
		val (nlp_term, prg_term) =
		    foldl (fn ((v, typ), (termN, termP)) =>
			      (NT.mk_nuprl_all_term typ (v, termN),
			       NT.mk_nuprl_all_term typ (v, termP)))
			  (nlpsimp, prgsimp)
			  (oargs @ args)
	    in SOME (vars, nlp_term, prg_term)
	    end
       else if opid = "function"
       then let val (t1, t2) = NT.dest_function nity
		val narg  = "x" ^ Int.toString (List.length args)
		val id    = A.newIdIdSet set narg
		val args' = args @ [(id, t1)]
		val set'  = IDS.add (set, id)
	    in genNormLocProg set' class t2 args'
	    end
       else NONE
    end

fun genAbstractionAndWF i value id n_exp exp schemeop programmable lvlop =
    let val eid    = if A.isInternalId value then value else id
	(*val _ = print ("gen-abs (" ^ Int.toString i ^ ")\n")*)
	val (n_ity, bounds) =
	    let val nb = getNbOfId eid
	    in case EN.getScheme eid (nb + 1) (getUnifEnv ()) of
		   NONE =>
		   (case schemeop of
			SOME scheme => scheme
		      | NONE => (print (":" ^ Int.toString i ^ ":" ^ value ^ ":" ^ id ^ ":" ^ eid);
				 raise EH.Impossible ("Couldn't find " ^ eid ^ "'s environment")))
		 | SOME scheme => toNuprlScheme scheme
	    end
	val _      = addDecFPList (getFreeParams [] exp)
	val parms  = getDecFP ()
	val n_prs  = map NT.mk_nuprl_variable_term parms
	val n_id   = mk_nuprl_user_term_lvlop id n_prs lvlop
	val n_abs  = NT.mk_nuprl_iabstraction_term n_id n_exp
	val n_typ  = NT.mk_nuprl_type_term "i'"
	val n_atyp = NT.mk_nuprl_valuealltype_term "i"
	val n_btyp = NT.mk_nuprl_valuealltype_term "i'"
	val n_mem  = NT.mk_nuprl_member_term n_ity n_id
	val nprms  = map (fn param =>
			     case getParam param of
				 SOME scheme => (param, schemeParamToHead scheme)
			       | NONE => raise EH.Impossible "")
			 parms
	fun toAllPrmsWf term =
	    foldr (fn ((param, (nity, nts)), n_term) =>
		      foldr (NT.mk_nuprl_uall_term n_typ)
			    (let val typ =
				      if NT.is_nuprl_type_term nity
				      then n_atyp
				      else nity
			      in NT.mk_nuprl_uall_term typ (param, n_term)
			      end (*NT.mk_nuprl_uall_term nity (param, n_term)*))
			    nts)
		  term
		  nprms
	fun toAllPrmsNlp term =
	    foldr (fn ((param, (nity, nts)), n_term) =>
		      ((*print ("[" ^ param ^ "]\n");
		       print (toStringTerm nity ^ "\n");*)
		       foldr (NT.mk_nuprl_all_term n_typ)
			     (let val typ =
				      if NT.is_nuprl_type_term nity
				      then n_atyp
				      else nity
			      in NT.mk_nuprl_all_term typ (param, n_term)
			      end)
			     nts))
		  term
		  nprms
	val n_thm  = let val all = foldr (NT.mk_nuprl_uall_term n_typ) n_mem bounds
		     in NT.mk_nuprl_itheorem_term (get_user_name id ^ "_wf") (toAllPrmsWf all)
		     end
	val set    = get_all_ids_list [n_id, n_ity]
	val n_nrms = case genNormLocProg set n_id n_ity [] of
			 SOME (vars, nlp_term, prg_term) =>
			 let val (n_all_N, n_all_P) =
				 foldr (fn (v, (tN, tP)) =>
					   if List.exists (fn x => x = v) vars
					   then (NT.mk_nuprl_all_term n_btyp (v, tN),
						 NT.mk_nuprl_all_term n_btyp (v, tP))
					   else (NT.mk_nuprl_all_term n_typ  (v, tN),
						 NT.mk_nuprl_all_term n_typ  (v, tP)))
				       (nlp_term, prg_term)
				       bounds
			     val all_nlp = toAllPrmsNlp n_all_N
			     val all_prg = toAllPrmsNlp n_all_P
			     val nlp_lem = NT.mk_nuprl_itheorem_term (get_user_name id ^ "_nlp") all_nlp
			     val prop    = NT.mk_nuprl_iproperty_term "extract_required" (NT.mk_nuprl_ibool_term true)
			     val cons    = NT.mk_nuprl_icons_cons_term prop NT.mk_nuprl_icons_nil_term
			     val ipropsn = NT.mk_nuprl_iinclude_properties_term cons nlp_lem
			     val prg_lem = NT.mk_nuprl_itheorem_term (get_user_name id ^ "_programmable") all_prg
			     val ipropsp = NT.mk_nuprl_iinclude_properties_term cons prg_lem
			 in if programmable
			    then [ipropsn, ipropsp]
			    else [ipropsn]
			 end
		       | NONE => []
	val _      = addOneId eid
	val _      = addUserFP id parms
    in [n_abs, n_thm] @ n_nrms
    end

fun genAbstractionAndWF' i value id n_exp exp =
    genAbstractionAndWF i value id n_exp exp NONE false NONE

fun genAbstractionAndWFDots i value id n_exp =
    genAbstractionAndWF' i value id n_exp (A.mk_new_dum_term A.DOTS_D "" [] [])

fun toNuprlCombComp n_exp n_exps bprior optop =
    let val list = foldr (fn (nt, list) => NT.mk_nuprl_cons_term nt list)
			 NT.mk_nuprl_nil_term
			 n_exps
	val set  = get_all_ids_list [n_exp, list]
	val id   = A.newIdIdSet set "n"
	val tid  = NT.mk_nuprl_variable_term id
	val sel  = NT.mk_nuprl_lambda_term id (NT.mk_nuprl_select_term tid list)
	(* NOTE: sel is the list of classes as a function. *)
	val fid  = A.newIdIdSet set "w"
	val sid  = A.newIdIdSet set "s"
	val tfid = NT.mk_nuprl_variable_term fid
	val tsid = NT.mk_nuprl_variable_term sid
	val lid  = A.newIdIdSet set "loc"
	val thd  = let val tlid = NT.mk_nuprl_variable_term lid
		   in NT.mk_nuprl_apply_term n_exp tlid
		   end
	val nt'  = #1 (foldr (fn (_, (nt, n)) =>
				 let val n_term = NT.mk_nuprl_natural_number_term n
				     val app1   = NT.mk_nuprl_apply_term tfid n_term
				     val app2   = NT.mk_nuprl_apply_term nt app1
				 in (app2, n + 1)
				 end)
			     (thd, 0)
			     n_exps)
	val app  = NT.mk_nuprl_apply_term nt' tsid
	val lam0 = NT.mk_nuprl_lambda_term sid app
	val lam1 = NT.mk_nuprl_lambda_term fid lam0
	val lam2 = NT.mk_nuprl_lambda_term lid lam1
    in if bprior
       then case optop of
		SOME opt  => NT.mk_nuprl_rec_comb_term lam2 sel opt
	      | NONE      => NT.mk_nuprl_rec_combined_loc_class_term lam2 sel
       else NT.mk_nuprl_combined_loc_class_term lam2 sel
    end

fun toLift f blift flift bconcat fconcat =
    if blift
    then if bconcat
	 then fconcat f
	 else flift f
    else f

fun preToNuprlCombComp n_exp n_exps bprior opt blift bconcat =
    let val (f0, f1, f2, f3) =
	    if bprior
	    then case opt of
		     SOME opt => (NT.mk_nuprl_rec_combined0_loc_class_opt_term opt,
				  NT.mk_nuprl_rec_combined1_loc_class_opt_term opt,
				  NT.mk_nuprl_rec_combined2_loc_class_opt_term opt,
				  NT.mk_nuprl_rec_combined3_loc_class_opt_term opt)
		   | NONE     => (NT.mk_nuprl_rec_combined0_loc_class_term,
				  NT.mk_nuprl_rec_combined1_loc_class_term,
				  NT.mk_nuprl_rec_combined2_loc_class_term,
				  NT.mk_nuprl_rec_combined3_loc_class_term)
	    else (NT.mk_nuprl_combined0_loc_class_term,
		  NT.mk_nuprl_combined1_loc_class_term,
		  NT.mk_nuprl_combined2_loc_class_term,
		  NT.mk_nuprl_combined3_loc_class_term)
	fun toLift' flift fconcat = toLift n_exp blift flift bconcat fconcat
	fun toLift0 () =
	    if bprior
	    then toLift' NT.mk_nuprl_lifting_loc1_term NT.mk_nuprl_concat_lifting_loc1_term
	    else toLift' NT.mk_nuprl_lifting_loc0_term NT.mk_nuprl_concat_lifting_loc0_term
	fun toLift1 () =
	    if bprior
	    then toLift' NT.mk_nuprl_lifting_loc2_term NT.mk_nuprl_concat_lifting_loc2_term
	    else toLift' NT.mk_nuprl_lifting_loc1_term NT.mk_nuprl_concat_lifting_loc1_term
	fun toLift2 () =
	    if bprior
	    then toLift' NT.mk_nuprl_lifting_loc3_term NT.mk_nuprl_concat_lifting_loc3_term
	    else toLift' NT.mk_nuprl_lifting_loc2_term NT.mk_nuprl_concat_lifting_loc2_term
	fun toLift3 () =
	    if bprior
	    then toLift' NT.mk_nuprl_lifting_loc4_term NT.mk_nuprl_concat_lifting_loc4_term
	    else toLift' NT.mk_nuprl_lifting_loc3_term NT.mk_nuprl_concat_lifting_loc3_term
	val n = List.length n_exps
	fun toLiftGen () =
	    if bprior
	    then let val nt = NT.mk_nuprl_natural_number_term (n + 1)
		 in toLift' (NT.mk_nuprl_lifting_loc_gen_term nt)
			    (NT.mk_nuprl_concat_lifting_loc_gen_term nt)
		 end
	    else let val nt = NT.mk_nuprl_natural_number_term n
		 in toLift' (NT.mk_nuprl_lifting_loc_gen_term nt)
			    (NT.mk_nuprl_concat_lifting_loc_gen_term nt)
		 end
    in case n_exps of
	   [] => f0 (toLift0 ())
	 | [ne] => f1 (toLift1 ()) ne
	 | [ne1, ne2] => f2 (toLift2 ()) ne1 ne2
	 | _ => toNuprlCombComp (toLiftGen ()) n_exps bprior opt
    end

fun transfun #" " = "\\ "
  | transfun x = Char.toString x

val transform = String.translate transfun

(* Transform a term into a nuprl term *)
fun toNuprlTerm (term as A.N {kind = (A.SCON, A.SCON_INT),    label, value, regions, parents, children = []}) =
    (case Int.fromString value of
	 NONE   => raise EH.Impossible A.wrongFormat
       | SOME n => NT.mk_nuprl_integer_term n)
  | toNuprlTerm (term as A.N {kind = (A.SCON, A.SCON_REAL),   label, value, regions, parents, children = []}) = raise EH.Unimplemented ""
  | toNuprlTerm (term as A.N {kind = (A.SCON, A.SCON_ATOM),   label, value, regions, parents, children = []}) = NT.mk_nuprl_token_term value
  | toNuprlTerm (term as A.N {kind = (A.SCON, A.SCON_ATOMS),  label, value, regions, parents, children = []}) = raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.SCON, A.SCON_STRING), label, value, regions, parents, children = []}) = raise EH.Impossible minStr

  (* ID *)
  | toNuprlTerm (term as A.N {kind = (A.ID, A.ID_VID),    label, value, regions, parents, children = []}) =
    (case value of
	 "true"  => NT.mk_nuprl_btrue_term
       | "false" => NT.mk_nuprl_bfalse_term
       | "nil"   => NT.mk_nuprl_nil_term
       | "="     =>
	 (case EN.getItyAtLab (getUnifEnv ()) label of
	      NONE => raise EH.Impossible "We shoud have generated a type for this equality"
	    | SOME ity => toDeq (E.stripIty ity))
       | _ =>
	 (case getArity value of
	      NONE => (* the id is not a constant, it is a user function *)
	      (case getUserFP value of
		   (* We sometimes have to add parameters to the argument list
		    * of a used function, if it does use the parameters. *)
		   NONE => NT.mk_nuprl_variable_term value
		 | SOME params =>
		   let val a_terms = map NT.mk_nuprl_variable_term params
		       val _ = addDecFPList params
		   in mk_nuprl_user_term value a_terms
		   end)
	    | SOME (idname, lst) =>
	      if isParam value
	      then if List.null lst
		   then NT.mk_nuprl_variable_term value
		   else raise EH.Impossible ""
	      else let val name =
			   case EN.getNameAtLab (getUnifEnv ()) label of
			       NONE => idname
			     | SOME name => name
		       val vars =
			   List.map (fn (id, args) =>
					let val set = IDS.addList (IDS.singleton name, args)
					in (A.newIdIdSet set id, args)
					end)
				    lst
		       val s_terms =
			   map (fn (id, args) =>
				   (args,
				    foldl (fn (arg, term) =>
					      let val t_arg = NT.mk_nuprl_variable_term arg
					      in NT.mk_nuprl_apply_term term t_arg
					      end)
					  (NT.mk_nuprl_variable_term id)
					  args))
			       vars
		       val term = NT.mk_nuprl_term (name, []) s_terms
		   in foldr (fn ((id, _), term) => NT.mk_nuprl_lambda_term id term)
			    term
			    vars
		   end))
  | toNuprlTerm (term as A.N {kind = (A.ID, A.ID_TYCON),  label, value, regions, parents, children = []}) = raise EH.Unimplemented ""
  | toNuprlTerm (term as A.N {kind = (A.ID, A.ID_TYVAR),  label, value, regions, parents, children = []}) = raise EH.Unimplemented ""

  (*(* ATPAT *)
  | toNuprlTerm (term as A.N {kind = (A.ATPAT, A.ATPAT_ID),    label, value, regions, parents, children = [i]})  = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.ATPAT, A.ATPAT_WILD),  label, value, regions, parents, children = []})   = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.ATPAT, A.ATPAT_SCON),  label, value, regions, parents, children = [sc]}) = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.ATPAT, A.ATPAT_LIST),  label, value, regions, parents, children})        = raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.ATPAT, A.ATPAT_PAREN), label, value, regions, parents, children = [x]})  = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.ATPAT, A.ATPAT_TUPLE), label, value, regions, parents, children})        = raise EH.TODO*)

  (*(* PAT *)
  | toNuprlTerm (term as A.N {kind = (A.PAT, A.PAT_TYPED), label, value, regions, parents, children = [pat, ty]})    = raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.PAT, A.PAT_CONS),  label, value, regions, parents, children = [pat1, pat2]}) = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.PAT, A.PAT_ATPAT), label, value, regions, parents, children = [x]})          = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.PAT, A.PAT_APP),   label, value, regions, parents, children = [f, x]})       = raise EH.TODO*)

  (*(* MRULE  *)
  | toNuprlTerm (term as A.N {kind = (A.MRULE, A.MRULE_C),  label, value, regions, parents, children = [a, e]}) = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.MRULE, A.MRULE_C2), label, value, regions, parents, children = [a, e]}) = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.MRULE, A.MRULE_M),  label, value, regions, parents, children = [p, e]}) = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.MRULE, A.MRULE_B),  label, value, regions, parents, children = [b, e]}) = raise EH.TODO

  (* MATCH *)
  | toNuprlTerm (term as A.N {kind = (A.MATCH, A.MATCH_C),  label, value, regions, parents, children}) = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.MATCH, A.MATCH_C2), label, value, regions, parents, children}) = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.MATCH, A.MATCH_M),  label, value, regions, parents, children}) = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.MATCH, A.MATCH_B),  label, value, regions, parents, children}) = raise EH.TODO*)

  (* EXP *)
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_OR),       label, value, regions, parents, children = [e1, e2]})      =
    let val nt1 = toNuprlTerm e1
	val nt2 = toNuprlTerm e2
    in NT.mk_nuprl_bor_term nt1 nt2
    end
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_AND),      label, value, regions, parents, children = [e1, e2]})      =
    let val nt1 = toNuprlTerm e1
	val nt2 = toNuprlTerm e2
    in NT.mk_nuprl_band_term nt1 nt2
    end
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_TYPED),    label, value, regions, parents, children = [e, t]})        = raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_LAMBDA),   label, value, regions, parents, children = [p, e]})        =
    let val id    = getIdPat 4 p
	val b     = addBound id
	val nterm = toNuprlTerm e
	val _     = rmBound b id
    in NT.mk_nuprl_lambda_term id nterm
    end
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_ITE),      label, value, regions, parents, children = [e1, e2, e3]})  =
    let val nt1 = toNuprlTerm e1
	val nt2 = toNuprlTerm e2
	val nt3 = toNuprlTerm e3
    in NT.mk_nuprl_ite_term nt1 nt2 nt3
    end
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_ATEXP),    label, value, regions, parents, children = [ae]})          = toNuprlTerm ae
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_APP),      label, value, regions, parents, children = [f, x]})        =
    let val (t, ts) = flattenExpApp f
	val nts = map toNuprlTerm (ts @ [x])
    in case t of
	   A.N {kind = (A.EXP, A.EXP_ATEXP), label, value, regions, parents, children = [atexp]} =>
	   (case atexp of
		A.N {kind = (A.ATEXP, A.ATEXP_ID), label, value, regions, parents, children = [v]} =>
		let val id   = A.getIdIdent v
		    val lab  = A.getLabel v
		in case getArity id of
		       NONE =>
		       foldl (fn (arg, func) => NT.mk_nuprl_apply_term func arg)
			     (toNuprlTerm v)
			     nts
		     | SOME (idname, lst) =>
		       if isParam id
		       then if List.null lst
			    then foldl (fn (arg, func) => NT.mk_nuprl_apply_term func arg)
				       (NT.mk_nuprl_variable_term id)
				       nts
			    else raise EH.Impossible ""
		       else let val n  = List.length lst
				val ln = List.length nts
				val (params, args, bound) =
				    if ln >= n
				    then let val list1 = List.take (nts, n)
					     val list2 = List.drop (nts, n)
					 in (list1, list2, [])
					 end
				    else let val vars = List.tabulate (n - ln, fn _ => A.newId ())
					     (*val vars = map (fn (id, _) => A.newIdId id) (List.drop (lst, ln))*)
					 in (nts @ (map NT.mk_nuprl_variable_term vars), [], vars)
					 end
				val b_terms =
				    map (fn ((id, prms), arg) =>
					    let val (bounds, body) = NT.dest_lambdas arg
						val m  = List.length prms
						val bl = List.length bounds
					    in if bl >= m
					       then let val l1 = List.take (bounds, m)
							val l2 = List.drop (bounds, m)
						    in (l1, NT.mk_nuprl_lambdas_term l2 body)
						    end
					       else let val vars = List.tabulate (m - bl, fn _ => A.newId ())
							val term = foldl (fn (arg, term) =>
									     NT.mk_nuprl_apply_term term (NT.mk_nuprl_variable_term arg))
									 body
									 vars
						    in (bounds @ vars, term)
						    end
					    end)
					(ListPair.zip (lst, params))
				val name = case EN.getNameAtLab (getUnifEnv ()) lab of
					       NONE => idname
					     | SOME name => name
				val app =
				    foldl (fn (arg, func) => NT.mk_nuprl_apply_term func arg)
					  (NT.mk_nuprl_term (name, []) b_terms)
					  args
				val lam =
				    foldr (fn (f, term) =>
					      NT.mk_nuprl_lambda_term f term)
					  app
					  bound
			    in lam
			    end handle Subscript => raise EH.Impossible ""
		end
	      | _ => foldl (fn (arg, func) => NT.mk_nuprl_apply_term func arg) (toNuprlTerm atexp) nts)
	 | _ => foldl (fn (arg, func) => NT.mk_nuprl_apply_term func arg) (toNuprlTerm t) nts
    end
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_OP),       label, value, regions, parents, children = [e1, e2]})      =
    let val nt1 = toNuprlTerm e1
	val nt2 = toNuprlTerm e2
    in case value of
	   "+"     => NT.mk_nuprl_add_term       nt1 nt2
	 | "-"     => NT.mk_nuprl_subtract_term  nt1 nt2
	 | "*"     => NT.mk_nuprl_multiply_term  nt1 nt2
	 | "/"     => NT.mk_nuprl_divide_term    nt1 nt2
	 | "="     => (case EN.getItyAtLab (getUnifEnv ()) label of
			   (* NOTE: this use of labels assumes that we never change
			    * the label of a EXP_OP. *)
			   NONE => raise EH.Impossible "We shoud have generated a type for this equality"
			 | SOME ity =>
			   let val deq = toDeq (E.stripIty ity)
			       val app = NT.mk_nuprl_apply_term deq nt1
			   in NT.mk_nuprl_apply_term app nt2
			   end)
	 | "=="    => raise EH.Unimplemented ""
	 | "<>"    => (case EN.getItyAtLab (getUnifEnv ()) label of
			   NONE => raise EH.Impossible "We shoud have generated a type for this equality"
			 | SOME ity =>
			   let val deq  = toDeq (E.stripIty ity)
			       val app1 = NT.mk_nuprl_apply_term deq  nt1
			       val app2 = NT.mk_nuprl_apply_term app1 nt2
			   in NT.mk_nuprl_bnot_term app2
			   end)
	 | "<="    => let val leq  = NT.mk_nuprl_lt_int_term nt1 nt2
			  val deq  = NT.mk_nuprl_int_deq_term
			  val app1 = NT.mk_nuprl_apply_term deq  nt1
			  val app2 = NT.mk_nuprl_apply_term app1 nt2
		      in NT.mk_nuprl_bor_term leq app2
		      end
	 | ">="    => let val leq  = NT.mk_nuprl_lt_int_term nt2 nt1
			  val deq  = NT.mk_nuprl_int_deq_term
			  val app1 = NT.mk_nuprl_apply_term deq  nt1
			  val app2 = NT.mk_nuprl_apply_term app1 nt2
		      in NT.mk_nuprl_bor_term leq app2
		      end
	 | "<"     => NT.mk_nuprl_lt_int_term           nt1 nt2
	 | ">"     => NT.mk_nuprl_lt_int_term           nt2 nt1
	 | "."     => NT.mk_nuprl_cons_term             nt1 nt2
	 | "++"    => NT.mk_nuprl_concat_term           nt1 nt2
	 | "@"     => NT.mk_nuprl_class_at_term         nt1 nt2
	 | "||"    => NT.mk_nuprl_parallel_class_term   nt1 nt2
	 | "until" => NT.mk_nuprl_until_class_term      nt1 nt2
	 | "?"     =>
	   if NT.opid_of_term nt1 = "primed-class"
	   then NT.mk_nuprl_primed_class_opt_term (NT.dest_primed_class nt1) nt2
	   else NT.mk_nuprl_class_opt_term nt1 nt2
	 | "??"    => NT.mk_nuprl_class_opt_class_term nt1 nt2
	 | _       =>
	   (case getArity value of
		NONE => (* the id is not a constant, it is a user function *)
		(case getUserFP value of
		     (* We sometimes have to add parameters to the argument list
		      * of a used function, if it does use the parameters. *)
		     NONE => NT.mk_nuprl_variable_term value
		   | SOME params =>
		     let val a_terms = map NT.mk_nuprl_variable_term params
			 val _       = addDecFPList params
			 val func    = mk_nuprl_user_term value a_terms
		     in NT.mk_nuprl_apply_term (NT.mk_nuprl_apply_term func nt1) nt2
		     end)
	      | _ => raise EH.Impossible "a constant or a parameter cannot be infix")
    end
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_LET),      label, value, regions, parents, children = [b, e]})        =
    (case b of
	 A.N {kind = (A.BIND, A.BIND_DEC), label, value, regions, parents, children = [f, a, e']} =>
	 let val id = A.getIdIdent f
	     val ne = toNuprlTerm e'
	     val b  = addBound id
	     val nt = toNuprlTerm e
	     val _  = rmBound b id
	 in NT.mk_nuprl_let_term id ne nt
	 end
       | A.N {kind = (A.BIND, A.BIND_PAT), label, value = v, regions, parents, children = [p, e']} =>
	 (case p of
	      A.N {kind = (A.PAT, A.PAT_ATPAT), label, value, regions, parents, children = [atpat]} =>
	      (case atpat of
		   A.N {kind = (A.ATPAT, A.ATPAT_TUPLE), label, value, regions, parents, children} =>
		   if List.length children <= 7
		   then let val vars = map (getIdPat 5) children
			    val ne   = toNuprlTerm e'
			    val bs   = map addBound vars
			    val nt   = toNuprlTerm e
			    val lst  = ListPair.zip (bs, vars)
			    val _    = app (fn (b, v) => rmBound b v) lst
			in NT.mk_nuprl_spreadn_term ne (vars, nt)
			end
		   else raise EH.Impossible minStr
		 | A.N {kind = (A.ATPAT, A.ATPAT_ID), label, value, regions, parents, children = [ident]} =>
		   let val id = A.getIdIdent ident
		       val ne = toNuprlTerm e'
		       val b  = addBound id
		       val nt = toNuprlTerm e
		       val _  = rmBound b id
		   in NT.mk_nuprl_let_term id ne nt
		   end
		 | atpat => (print (A.toString atpat ^ "\n"); raise EH.Impossible A.wrongFormat))
	    | pat => (print (A.toString pat ^ "\n"); raise EH.Impossible minStr))
       | _ => raise EH.Impossible minStr)
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_LETR),     label, value, regions, parents, children = [b, e]})        =
    (case b of
	 A.N {kind = (A.BIND, A.BIND_DEC), label, value, regions, parents, children = [f, a, e']} =>
	 let val fid = A.getIdIdent f
	     val b   = addBound fid
	     val ne  = toNuprlTerm e'
	     val nt  = toNuprlTerm e
	     val _   = rmBound b fid
	 in if NT.is_nuprl_lambda_term ne
	    then let val (x, B) = NT.dest_lambda ne
		     (**)
		     (* val y    = A.newId ()
		      * val t_y  = NT.mk_nuprl_variable_term y
		      * (* rind: letrec f x = B in f y *)
		      * val rind = NT.mk_nuprl_rec_ind_term t_y (fid, x, B)
		      * (* lam1: \f. e *)
		      * val lam1 = NT.mk_nuprl_lambda_term fid nt
		      * (* lam2: \y. letrec f x = B in f y *)
		      * val lam2 = NT.mk_nuprl_lambda_term y rind
		      * (* term: (\f. e) (\y. letrec f x = B in f y) *)
		      * val term = NT.mk_nuprl_apply_term lam1 lam2*)
		     (**)
		     (* gind: letrec f x = B in f *)
		     val gind = NT.mk_nuprl_genrec_term (x, fid, B)
		     val term =
			 if NT.is_nuprl_variable_term nt
			    andalso
			    NT.dest_variable nt = fid
			 then gind
			 else let val lam = NT.mk_nuprl_lambda_term fid nt
			      (* lam: \f. e *)
			      (* term: (\f. e) (letrec f x = B in f) *)
			      in NT.mk_nuprl_apply_term lam gind
			      end
		 in term
		 end
	      else (* if NT.is_nuprl_simple_comb1_term ne
		    * then let val (x, f, class) = NT.dest_simple_comb1 ne
		    *      in if is_prior_id class fid
		    *         then NT.mk_nuprl_rec_combined0_class_term
		    *         else
		    *      end
		    * else raise EH.Unimplemented ""
		    * else *)
		  (print (A.toString e'); raise EH.Impossible "recursive definition is not a function")
	 end
       | _ => raise EH.Impossible minStr)
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_WHERE),    label, value, regions, parents, children = [e, b]})        =
    raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_BINDING),  label, value, regions, parents, children = [e1, p, e2]})   =
    (case A.getIdPat p of
	 SOME ident =>
	 let val nt1 = toNuprlTerm e1
	     val id  = A.getIdIdent ident
	     val b   = addBound id
	     val nt2 = toNuprlTerm e2
	     val _   = rmBound b id
	 in NT.mk_nuprl_bind_class_term nt1 (id, nt2)
	 end
       | NONE => (print (A.toString p ^ "\n"); raise EH.Impossible minStr))
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_MBIND),    label, value, regions, parents, children = [e1, e2]})      = raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_COMP),     label, value, regions, parents, children = (exp :: exps)}) =
    let val nt  = toNuprlTerm exp
	val nts = map toNuprlTerm exps
	val bp  = String.isSubstring "P" value
	val bf  = String.isSubstring "F" value
	val bc  = String.isSubstring "C" value
	val bo  = String.isSubstring "O" value
	val (nts', opt) =
	    if bo andalso not (List.null nts)
	    then let val rev    = List.rev nts
		     val last   = List.hd rev
		     val last_e = List.last exps
		     val firsts = List.rev (List.tl rev)
		 in if A.isExpBag last_e
		    then let val id  = A.newIdIdSet (get_all_ids_list [last]) "l"
			     val lam = NT.mk_nuprl_lambda_term id last
			 in (firsts, SOME lam)
			 end
		    else (firsts, SOME last)
		 end
	    else (nts, NONE)
    in preToNuprlCombComp nt nts' bp opt bf bc
    end
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_QUOT),     label, value, regions, parents, children = [e]})           = toNuprlTerm e
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_CASE),     label, value, regions, parents, children = [e, m]})        =
    let val e_term = toNuprlTerm e
    in case A.caseMatch m of
	   A.CASE_LIST (exp1, (id1, id2, exp2)) =>
	   let val nt1 = toNuprlTerm exp1
	       val b1  = addBound id1
	       val b2  = addBound id2
	       val nt2 = toNuprlTerm exp2
	       val _   = rmBound b1 id1
	       val _   = rmBound b2 id2
	       val t   = (id1, id2, A.newId (), nt2)
	   in NT.mk_nuprl_list_ind_term e_term nt1 t
	   end
	 | A.CASE_BAG (NONE, (pat, exp1), exp2) =>
	   let val b     = NT.is_nuprl_variable_term e_term
	       val mid   = if b then NT.dest_variable e_term else A.newId ()
	       val tmid  = NT.mk_nuprl_variable_term mid
	       val app   =
		   case A.getIdPat pat of
		       SOME id =>
		       let val lam = NT.mk_nuprl_lambda_term (A.getIdIdent id) (toNuprlTerm exp1)
		       in NT.mk_nuprl_apply_term lam (NT.mk_nuprl_bag_only_term tmid)
		       end
		     | NONE => raise EH.Impossible "We have transformed the pat of a msg casepat into an id."
	       val bsize = NT.mk_nuprl_bag_size_term tmid
	       val bone  = NT.mk_nuprl_eq_int_term bsize (NT.mk_nuprl_integer_term 1)
	       val ite   = NT.mk_nuprl_ite_term bone app (toNuprlTerm exp2)
	   in if b
	      then ite
	      else let val lam = NT.mk_nuprl_lambda_term mid ite
		   in NT.mk_nuprl_apply_term lam e_term
		   end
	   end
	 | A.CASE_BAG (SOME exp1, (pat, exp2), exp3) =>
	   let val b     = NT.is_nuprl_variable_term e_term
	       val mid   = if b then NT.dest_variable e_term else A.newId ()
	       val tmid  = NT.mk_nuprl_variable_term mid
	       val bnull = NT.mk_nuprl_bag_null_term tmid
	       val app   =
		   case A.getIdPat pat of
		       SOME id =>
		       let val lam = NT.mk_nuprl_lambda_term (A.getIdIdent id) (toNuprlTerm exp2)
		       in NT.mk_nuprl_apply_term lam (NT.mk_nuprl_bag_only_term tmid)
		       end
		     | NONE => raise EH.Impossible "We have transformed the pat of a msg casepat into an id."
	       val bsize = NT.mk_nuprl_bag_size_term tmid
	       val bone  = NT.mk_nuprl_eq_int_term bsize (NT.mk_nuprl_integer_term 1)
	       val ite1  = NT.mk_nuprl_ite_term bone app (toNuprlTerm exp3)
	       val ite2  = NT.mk_nuprl_ite_term bnull (toNuprlTerm exp1) ite1
	   in if b
	      then ite2
	      else let val lam = NT.mk_nuprl_lambda_term mid ite2
		   in NT.mk_nuprl_apply_term lam e_term
		   end
	   end
	 | A.CASE_MSG (lst, exp) =>
	   let val b     = NT.is_nuprl_variable_term e_term
	       val mid   = if b then NT.dest_variable e_term else A.newId ()
	       val tmid  = NT.mk_nuprl_variable_term mid
	       val t_hdr = NT.mk_nuprl_msg_header_term tmid
	       val t_typ = NT.mk_nuprl_msg_type_term   tmid
	       val t_bdy = NT.mk_nuprl_msg_body_term   tmid
	       val t_exp = toNuprlTerm exp
	       val nid   = A.newIdIdSet (get_all_ids_list [t_exp]) "_"
	       val lexp  = NT.mk_nuprl_lambda_term nid t_exp
	       val term  =
		   foldr (fn ((atom, pat, exp, labM), term) =>
			     case (A.getIdPat pat, atom) of
				 (SOME id, A.N {kind = (A.ATOMS, A.ATOMS_ATOMS), label, value, regions, parents, children = []}) =>
				 let val n_atoms = toNuprlListAtoms (A.splitAtoms value)
				     val n_eq1   = NT.mk_nuprl_name_eq_term t_hdr n_atoms
				     val n_ity   = case EN.getItyAtLab (getUnifEnv ()) labM of
						       NONE => raise EH.Impossible "we should have generated a type for each pattern message"
						     | SOME ity => toNuprlIty 3 (E.stripIty ity) BTV.empty E.emptyETVS
				     val n_eq2   = NT.mk_nuprl_msg_has_type_term tmid n_ity
				     val n_eq    = NT.mk_nuprl_band_term n_eq1 n_eq2
				     val lam     = NT.mk_nuprl_lambda_term (A.getIdIdent id) (toNuprlTerm exp)
				 in NT.mk_nuprl_ite_term n_eq lam term
				 end
			       | (NONE, _) => raise EH.Impossible "We have transformed the pat of a msg casepat into an id."
			       | _ => raise EH.Impossible "We have transformed the atoms into the slanted quoted forms.")
			 lexp
			 lst
	       val app = NT.mk_nuprl_apply_term term t_bdy
	   in if b
	      then app
	      else let val lam = NT.mk_nuprl_lambda_term mid app
		   in NT.mk_nuprl_apply_term lam e_term
		   end
	   end
	 | A.CASE_OTH => (print (A.toString m ^ "\n" ^ A.export m); raise EH.Unimplemented "")
    end
(*if A.isMatchBag m
       then raise EH.Unimplemented ""
       else if A.isMatchMsg m
       then let val b     = NT.is_nuprl_variable_term e_term
		val mid   = if b then NT.dest_variable e_term else A.newId ()
		val tmid  = NT.mk_nuprl_variable_term mid
		val t_hdr = NT.mk_nuprl_msg_header_term tmid
		val t_typ = NT.mk_nuprl_msg_type_term   tmid
		val t_bdy = NT.mk_nuprl_msg_body_term   tmid
	    in case m of
		   A.N {kind = (A.MATCH, A.MATCH_M), label, value, regions, parents, children} =>
		   if List.null children
		   then raise EH.Impossible "a case expression should have at least one branch"
		   else let val rev    = List.rev children
			    val last   = List.hd rev
			    val firsts = List.rev (List.tl rev)
			    val lexp   =
				case last of
				    A.N {kind = (A.MRULE, A.MRULE_M), label, value, regions, parents, children = [casepat, exp]} =>
				    let val id = A.newId ()
				    in NT.mk_nuprl_lambda_term id (toNuprlTerm exp)
				    end
				  (* NOTE: the casepat must be a wildcard *)
				  | _ => raise EH.Impossible A.wrongFormat
			    val term   =
				foldr (fn (A.N {kind = (A.MRULE, A.MRULE_M), label, value, regions, parents, children = [casepat, exp]}, term) =>
					  (case casepat of
					       A.N {kind = (A.CASEPAT, A.CASEPAT_MSG), label = labM, value, regions, parents, children = [atoms, pat]} =>
					       (case A.getIdPat pat of
						    SOME id =>
						    (case atoms of
							 A.N {kind = (A.ATOMS, A.ATOMS_ATOMS), label, value, regions, parents, children = []} =>
							 let val n_atoms = toNuprlListAtoms (tokenize value)
							     val n_eq1   = NT.mk_nuprl_name_eq_term t_hdr n_atoms
							     val n_ity   = case EN.getItyAtLab (getUnifEnv ()) labM of
									       NONE => raise EH.Impossible "we should have generated a type for each pattern message"
									     | SOME ity => toNuprlIty 3 (E.stripIty ity) BTV.empty E.emptyETVS
							     val n_eq2   = NT.mk_nuprl_msg_has_type_term tmid n_ity
							     val n_eq    = NT.mk_nuprl_band_term n_eq1 n_eq2
							     val lam     = NT.mk_nuprl_lambda_term (A.getIdIdent id) (toNuprlTerm exp)
							 in NT.mk_nuprl_ite_term n_eq lam term
							 end
						       | _ => raise EH.Impossible "We have transformed the atoms into the slanted quoted forms.")
						  | NONE => raise EH.Impossible "We have transformed the pat of a msg casepat into an id.")
					     | _ => raise EH.Impossible "We have already checked that it is a msg case.")
					| _ => raise EH.Impossible A.wrongFormat)
				      lexp
				      firsts
			    val app = NT.mk_nuprl_apply_term term t_bdy
			in if b
			   then app
			   else let val lam = NT.mk_nuprl_lambda_term mid app
				in NT.mk_nuprl_apply_term lam e_term
				end
			end
		 | _ => raise EH.Impossible A.wrongFormat
	    end
       else if A.isMatchList m
       then case m of
		A.N {kind = (A.MATCH, A.MATCH_M), label, value, regions, parents, children = [mrule1, mrule2]} =>
		(case (mrule1, mrule2) of
		     (A.N {kind = (A.MRULE, A.MRULE_M), label = _, value = _, regions = _, parents = _, children = [casepat1, exp1]},
		      A.N {kind = (A.MRULE, A.MRULE_M), label = _, value = _, regions = _, parents = _, children = [casepat2, exp2]}) =>
		     (case (casepat1, casepat2) of
			  (A.N {kind = (A.CASEPAT, A.CASEPAT_PAT), label = _, value = _, regions = _, parents = _, children = [pat1]},
			   A.N {kind = (A.CASEPAT, A.CASEPAT_PAT), label = _, value = _, regions = _, parents = _, children = [pat2]}) =>
			  let val (patCons, eCons, eNil) =
				  if A.getKind pat1 = (A.PAT, A.PAT_CONS)
				  then (pat1, toNuprlTerm exp1, toNuprlTerm exp2)
				  else (pat2, toNuprlTerm exp2, toNuprlTerm exp1)
			  in case patCons of
				 A.N {kind = (A.PAT, A.PAT_CONS), label, value, regions, parents, children = [pat1, pat2]} =>
				 (case (pat1, pat2) of
				      (A.N {kind = (A.PAT, A.PAT_ATPAT), label = _, value = _, regions = _, parents = _, children = [atpat1]},
				       A.N {kind = (A.PAT, A.PAT_ATPAT), label = _, value = _, regions = _, parents = _, children = [atpat2]}) =>
				      (case (atpat1, atpat2) of
					   (A.N {kind = (A.ATPAT, A.ATPAT_ID), label = _, value = _, regions = _, parents = _, children = [id1]},
					    A.N {kind = (A.ATPAT, A.ATPAT_ID), label = _, value = _, regions = _, parents = _, children = [id2]}) =>
					   let val i1 = A.getIdIdent id1
					       val i2 = A.getIdIdent id2
					       val id = A.newId ()
					   in NT.mk_nuprl_list_ind_term e_term eNil (i1, i2, id, eCons)
					   end
					 | _ => raise EH.Impossible "")
				    | _ => raise EH.Impossible "")
			       | _ => raise EH.Impossible ""
			  end
			| _ => raise EH.Impossible "")
		   | _ => raise EH.Impossible "")
	      | _ => raise EH.Impossible ""
       else raise EH.Unimplemented ""
    end*)
  (*| toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_CASEMSG),  label, value, regions, parents, children = [e, i, m]})     =
    let val idv   = A.getIdIdent i
	val ida   = A.newId ()
	val ta    = NT.mk_nuprl_variable_term ida
	val n_exp = toNuprlTerm e
    in case m of
	   A.N {kind = (A.MATCH, A.MATCH_C), label, value, regions, parents, children} =>
	   if List.null children
	   then raise EH.Impossible "a case expression should have at least one branch"
	   else let val rev    = List.rev children
		    val last   = List.hd rev
		    val firsts = List.rev (List.tl rev)
		    val lexp   =
			case last of
			    A.N {kind = (A.MRULE, A.MRULE_C), label, value, regions, parents, children = [_, exp]} => exp
			  (* NOTE: the atoms part must be a wildcard *)
			  | _ => raise EH.Impossible A.wrongFormat
		    val term   =
			foldr (fn (A.N {kind = (A.MRULE, A.MRULE_C), label, value, regions, parents, children = [atoms, exp]}, term) =>
				  let val n_exp = toNuprlTerm exp
				  in case atoms of
					 A.N {kind = (A.ATOMS, A.ATOMS_ATOMS), label, value, regions, parents, children = []} =>
					 let val n_atoms = toNuprlListAtoms (tokenize value)
					     val n_eq    = NT.mk_nuprl_equal_term ta n_atoms NT.mk_nuprl_atom_term
					     val lam     = NT.mk_nuprl_lambda_term idv n_exp
					 in NT.mk_nuprl_ite_term n_eq lam term
					 end
				       | _ => raise EH.Impossible ""
				  end
				| _ => raise EH.Impossible A.wrongFormat)
			      (toNuprlTerm lexp)
			      firsts
		    val lam    = NT.mk_nuprl_lambda_term ida term
		    val tpi1   = NT.mk_nuprl_apply_term NT.mk_nuprl_pi1_term n_exp
		    val tpi3   = NT.mk_nuprl_apply_term NT.mk_nuprl_pi2_term (NT.mk_nuprl_apply_term NT.mk_nuprl_pi2_term n_exp)
		    val app1   = NT.mk_nuprl_apply_term lam tpi1
		    val app2   = NT.mk_nuprl_apply_term app1 tpi3
		in app2
		end
	 | _ => raise EH.Impossible A.wrongFormat
    end
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_CASEMSG2), label, value, regions, parents, children = [e, m]})        = raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_CASEBAG),  label, value, regions, parents, children = [e, m]})        = raise EH.TODO
  | toNuprlTerm (term as A.N {kind = (A.EXP, A.EXP_MATCH),    label, value, regions, parents, children = [e, m]})        =
    if A.isMatchList m
    then let val e_term = toNuprlTerm e
	 in case m of
		A.N {kind = (A.MATCH, A.MATCH_M), label, value, regions, parents, children = [mrule1, mrule2]} =>
		(case (mrule1, mrule2) of
		     (A.N {kind = (A.MRULE, A.MRULE_M), label = _, value = _, regions = _, parents = _, children = [pat1, exp1]},
		      A.N {kind = (A.MRULE, A.MRULE_M), label = _, value = _, regions = _, parents = _, children = [pat2, exp2]}) =>
		     let val e1_term = toNuprlTerm exp1
			 val e2_term = toNuprlTerm exp2
		     (* NOTE: we know that pat1 is nil *)
		     in case pat2 of
			    A.N {kind = (A.PAT, A.PAT_CONS), label, value, regions, parents, children = [pat1, pat2]} =>
			    (case (pat1, pat2) of
				 (A.N {kind = (A.PAT, A.PAT_ATPAT), label = _, value = _, regions = _, parents = _, children = [atpat1]},
				  A.N {kind = (A.PAT, A.PAT_ATPAT), label = _, value = _, regions = _, parents = _, children = [atpat2]}) =>
				 (case (atpat1, atpat2) of
				      (A.N {kind = (A.ATPAT, A.ATPAT_ID), label = _, value = _, regions = _, parents = _, children = [id1]},
				       A.N {kind = (A.ATPAT, A.ATPAT_ID), label = _, value = _, regions = _, parents = _, children = [id2]}) =>
				      let val i1 = A.getIdIdent id1
					  val i2 = A.getIdIdent id2
					  val id = A.newId ()
				      in NT.mk_nuprl_list_ind_term e_term e1_term (i1, i2, id, e2_term)
				      end
				    | _ => raise EH.Impossible "")
			       | _ => raise EH.Impossible "")
			  | _ => raise EH.Impossible ""
		     end
		   | _ => raise EH.Impossible "")
	      | _ => raise EH.Impossible ""
	 end
    else raise EH.Impossible minStr*)

  (* ATEXP *)
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_ID),      label, value, regions, parents, children = [i]})      = toNuprlTerm i
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_SCON),    label, value, regions, parents, children = [sc]})     = toNuprlTerm sc
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_TUPLE),   label, value, regions, parents, children = []})       = NT.mk_nuprl_it_term
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_TUPLE),   label, value, regions, parents, children = [e1, e2]}) =
    let val nt1 = toNuprlTerm e1
	val nt2 = toNuprlTerm e2
    in NT.mk_nuprl_pair_term nt1 nt2
    end
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_TUPLE),   label, value, regions, parents, children})            = raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_PAREN),   label, value, regions, parents, children = [x]})      = toNuprlTerm x
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_LIST),    label, value, regions, parents, children})            = raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_BAG),     label, value, regions, parents, children})            =
    (case children of
	 [x] => NT.mk_nuprl_single_bag_term (toNuprlTerm x)
       | _   => foldr (fn (x, term) => NT.mk_nuprl_cons_bag_term (toNuprlTerm x) term)
		      NT.mk_nuprl_empty_bag_term
		      children)
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_PRIOR),   label, value, regions, parents, children = [e]})      = NT.mk_nuprl_primed_class_term (toNuprlTerm e)
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_BASE),    label, value, regions, parents, children = [e1, e2]}) =
    let val e_term1 = toNuprlTerm e1
	val e_term2 = toNuprlTerm e2
    in NT.mk_nuprl_general_base_class_term e_term1 e_term2
    end
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_ABASE),   label, value, regions, parents, children = [a, e]}) =
    let val hdrs =
	    case a of
		A.N {kind = (A.ATOMSL, A.ATOMSL_S), label, value, regions, parents, children} =>
		foldr (fn (atoms, list) => NT.mk_nuprl_cons_term (toAtomList atoms) list)
		      NT.mk_nuprl_nil_term
		      children
	      | _ => raise EH.Impossible ""
	val locs = toNuprlTerm e
    in NT.mk_nuprl_base_locs_headers_term hdrs locs
    end
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_ANY),     label, value, regions, parents, children = [e]})      = raise EH.Impossible "cannot convert an any term"
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_MSG),     label, value, regions, parents, children = [a, e]})   =
    (* TODO: We need type information here, if we know that the piece of code
     * is typable and if we got the generated environment then, we can just
     * lookup the type associated to the atoms of the message.
     * We can't actually do that because we can have a type message of the form
     * (``foo`` : 'a List) and the type of the message we're looking at is of
     * type Int List. *)
    let val nta = toAtomList a
	val nte = toNuprlTerm e
    in case EN.getItyAtLab (getUnifEnv ()) label of
	   NONE     => raise EH.Impossible "We should have generated a type for this message"
	 | SOME ity =>
	   let val n_ity = toNuprlIty 4 (E.stripIty ity) BTV.empty E.emptyETVS
	   in NT.mk_nuprl_make_msg_term nta n_ity nte
	   end
    end
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_GET), label, value, regions, parents, children = [a]}) =
     raise EH.Unimplemented ""
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_ONCE), label, value, regions, parents, children = [e]}) =
    let val nte = toNuprlTerm e
    in NT.mk_nuprl_once_class_term nte
    end
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_SENDOC), label, value, regions, parents, children = [e]}) =
    let val nte = toNuprlTerm e
    in NT.mk_nuprl_send_once_loc_class_term nte
    end
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_ONLOC), label, value, regions, parents, children = [e]}) =
    let val nte = toNuprlTerm e
    in NT.mk_nuprl_on_loc_class_term nte
    end
  | toNuprlTerm (term as A.N {kind = (A.ATEXP, A.ATEXP_MINUS), label, value, regions, parents, children = [e]}) =
    let val nte = toNuprlTerm e
    in NT.mk_nuprl_minus_term nte
    end

  (*(* ATOMS *)
  | toNuprlTerm (term as A.N {kind = (A.ATOMS, A.ATOMS_ATOM),  label, value, regions, parents, children = []}) = raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.ATOMS, A.ATOMS_ATOMS), label, value, regions, parents, children = []}) = term
  | toNuprlTerm (term as A.N {kind = (A.ATOMS, A.ATOMS_LIST),  label, value, regions, parents, children = []}) = raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.ATOMS, A.ATOMS_WILD),  label, value, regions, parents, children = []}) = raise EH.TODO*)

  (* PARAM *)
  | toNuprlTerm (term as A.N {kind = (A.PARAM, A.PARAM_P),  label, value, regions, parents, children}) = raise EH.Unimplemented ""

  (* TYPEVARSEQ *)
  | toNuprlTerm (term as A.N {kind = (A.TYPEVARSEQ, A.TYPEVARSEQ_ONE), label, value, regions, parents, children = [t]}) = raise EH.Unimplemented ""
  | toNuprlTerm (term as A.N {kind = (A.TYPEVARSEQ, A.TYPEVARSEQ_SEQ), label, value, regions, parents, children})       = raise EH.Unimplemented ""
  | toNuprlTerm (term as A.N {kind = (A.TYPEVARSEQ, A.TYPEVARSEQ_EM),  label, value, regions, parents, children = []})  = raise EH.Unimplemented ""

  (* TYPESEQ *)
  | toNuprlTerm (term as A.N {kind = (A.TYPESEQ, A.TYPESEQ_ONE), label, value, regions, parents, children = [t]}) = raise EH.Unimplemented ""
  | toNuprlTerm (term as A.N {kind = (A.TYPESEQ, A.TYPESEQ_SEQ), label, value, regions, parents, children})       = raise EH.Unimplemented ""
  | toNuprlTerm (term as A.N {kind = (A.TYPESEQ, A.TYPESEQ_EM),  label, value, regions, parents, children = []})  = raise EH.Unimplemented ""

  (* DOTS *)
  | toNuprlTerm (term as A.N {kind = (A.DOTS, A.DOTS_D), label, value, regions, parents, children}) = raise EH.Unimplemented ""

  (*(* BIND *)
  | toNuprlTerm (term as A.N {kind = (A.BIND, A.BIND_DEC),  label, value, regions, parents, children = [f, a, e]})       = raise EH.Impossible "this should be dealt with in the let-declarations and expressions"
  | toNuprlTerm (term as A.N {kind = (A.BIND, A.BIND_TDEC), label, value, regions, parents, children = [f, a, t, e]})    = raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.BIND, A.BIND_PAT),  label, value, regions, parents, children = [p, e]})          = raise EH.Impossible minStr
  | toNuprlTerm (term as A.N {kind = (A.BIND, A.BIND_IOP),  label, value, regions, parents, children = [id, a1, a2, e]}) = raise EH.Impossible minStr*)

  (* wrong format *)
  | toNuprlTerm term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

fun toNuprlAtomElt atm typ (term as A.N {kind = (A.ATOMELT, A.ATOMELT_BASE), label, value, regions, parents, children = [ident]}) =
    let val id   = A.getIdIdent ident
	val term = NT.mk_nuprl_base_headers_msg_val_term atm typ
	val res  = genAbstractionAndWFDots 1 value id term
	val _    = resetDecFP ()
    in res
    end
  | toNuprlAtomElt atm typ (term as A.N {kind = (A.ATOMELT, A.ATOMELT_BASEL), label, value, regions, parents, children = [ident, exp]}) =
    let val id   = A.getIdIdent ident
	val ne   = toNuprlTerm exp
	val term = NT.mk_nuprl_base_headers_msg_val_loc_term atm ne typ
	val res  = genAbstractionAndWF' 2 value id term exp
	val _    = resetDecFP ()
    in res
    end
  | toNuprlAtomElt atm typ (term as A.N {kind = (A.ATOMELT, A.ATOMELT_MSG), label, value, regions, parents, children = [ident]}) =
    let val id   = A.getIdIdent ident
	val set  = get_all_ids_list [atm, typ]
	val v    = A.newIdIdSet set "z"
	val tv   = NT.mk_nuprl_variable_term v
	(* msg: <atom, type, value> *)
	val msg  = NT.mk_nuprl_make_msg_term atm typ tv
	(* msg: \value. <atom, type, value> *)
	val term = NT.mk_nuprl_lambda_term v msg
	val res  = genAbstractionAndWFDots 3 value id term
	val _    = resetDecFP ()
    in res
    end
  | toNuprlAtomElt atm typ (term as A.N {kind = (A.ATOMELT, A.ATOMELT_SEND), label, value, regions, parents, children = [ident]}) =
    let val id   = A.getIdIdent ident
	val set  = get_all_ids_list [atm, typ]
	val v    = A.newIdIdSet set "z"
	val l    = A.newIdIdSet set "l"
	val tv   = NT.mk_nuprl_variable_term v
	val tl   = NT.mk_nuprl_variable_term l
	(* msg: <atom, type, value> *)
	val msg  = NT.mk_nuprl_make_msg_term atm typ tv
	(* pair: <loc, <atom, type, value>> *)
	val pair = NT.mk_nuprl_pair_term tl msg
	(* lam: \value. <loc, <atom, type, value>> *)
	val lam  = NT.mk_nuprl_lambda_term v pair
	(* term: \loc. \value. <loc, <atom, type, value>> *)
	val term = NT.mk_nuprl_lambda_term l lam
	val res  = genAbstractionAndWFDots 4 value id term
	val _    = resetDecFP ()
    in res
    end
  | toNuprlAtomElt atm typ (term as A.N {kind = (A.ATOMELT, A.ATOMELT_BCAST), label, value, regions, parents, children = [ident]}) =
    let val id   = A.getIdIdent ident
	val set  = get_all_ids_list [atm, typ]
	val v    = A.newIdIdSet set "z"
	val l    = A.newIdIdSet set "l"
	val ls   = A.newIdIdSet set "locs"
	val tv   = NT.mk_nuprl_variable_term v
	val tl   = NT.mk_nuprl_variable_term l
	val tls  = NT.mk_nuprl_variable_term ls
	(* msg: <atom, type, value> *)
	val msg  = NT.mk_nuprl_make_msg_term atm typ tv
	(* pair: <loc, <atom, type, value>> *)
	val pair = NT.mk_nuprl_pair_term tl msg
	(* llam:  \loc. <loc, <atom, type, value>> *)
	val llam = NT.mk_nuprl_lambda_term l pair
	(* bmap: bag-map(\loc. <loc, <atom, type, value>>,locs) *)
	val bmap = NT.mk_nuprl_bag_map_term llam tls
	(* lam: \value. bag-map(\loc. <loc, <atom, type, value>>,locs) *)
	val lam  = NT.mk_nuprl_lambda_term v bmap
	(* term: \locs. \value. bag-map(\loc. <loc, <atom, type, value>>,locs) *)
	val term = NT.mk_nuprl_lambda_term ls lam
	val res  = genAbstractionAndWFDots 5 value id term
	val _    = resetDecFP ()
    in res
    end
  | toNuprlAtomElt atm typ term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

(* send = \l. \v. (l, MSG(``hdr``, v)) *)
(* broadcast = \locs. \v. map (\l. (l, MSG(``hdr``, v))) locs *)

fun toNuprlAtomElts atm typ (term as A.N {kind = (A.ATOMELTS, A.ATOMELTS_ELTS), label, value, regions, parents, children}) =
    foldr (fn (elt, decs) => (toNuprlAtomElt atm typ elt) @ decs)
	  []
	  children
  | toNuprlAtomElts atm typ term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

(* TATOM *)
fun toNuprlTAtom (term as A.N {kind = (A.TATOM, A.TATOM_C), label, value, regions, parents, children = [i, a, t]}) =
    let val id   = A.getIdIdent i
	val atm  = toAtomList a
	val typ  =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		NONE     => raise EH.Impossible "We should have generated a type for this message"
	      | SOME ity => toNuprlIty 5 (E.stripIty ity) BTV.empty E.emptyETVS
	val _    = addHeader (value, atm)
	val term = NT.mk_nuprl_base_headers_msg_val_term atm typ
	val res  = genAbstractionAndWFDots 6 value id term
	val _    = resetDecFP ()
    in res
    end
  | toNuprlTAtom (term as A.N {kind = (A.TATOM, A.TATOM_L), label, value, regions, parents, children = [i, e, a, t]}) =
    let val id   = A.getIdIdent i
	val ne   = toNuprlTerm e
	val atm  = toAtomList a
	val typ  =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		NONE     => raise EH.Impossible "We should have generated a type for this message"
	      | SOME ity => toNuprlIty 6 (E.stripIty ity) BTV.empty E.emptyETVS
	val _    = addHeader (value, atm)
	val term = NT.mk_nuprl_base_headers_msg_val_loc_term atm ne typ
	val res  = genAbstractionAndWF' 7 value id term e
	val _    = resetDecFP ()
    in res
    end
  | toNuprlTAtom (term as A.N {kind = (A.TATOM, A.TATOM_G), label, value, regions, parents, children = [atoms, typ, elts]}) =
    let val atm  = toAtomList atoms
	val typ  =
	    case EN.getItyAtLab (getUnifEnv ()) label of
		NONE     => raise EH.Impossible "We should have generated a type for this message"
	      | SOME ity => toNuprlIty 7 (E.stripIty ity) BTV.empty E.emptyETVS
	val _    = addHeader (value, atm)
    in toNuprlAtomElts atm typ elts
    end
  | toNuprlTAtom term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

(* DECS *)
fun toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_LET), label, value, regions, parents, children = [b]}) =
    (case b of
	 A.N {kind = (A.BIND, A.BIND_DEC),  label, value, regions, parents, children = [f, a, e]} =>
	 if A.termIsEmParam a
	 then let val id    = A.getIdIdent f
		  val n_exp = toNuprlTerm e
		  val res   = genAbstractionAndWF' 8 value id n_exp e
		  val _     = resetDecFP ()
	      in res
	      end
	 else raise EH.Impossible minStr
       | _ => raise EH.Impossible minStr)
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_LETR),    label, value, regions, parents, children = [b]})              = raise EH.Impossible minStr
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_CONS),    label, value, regions, parents, children = [i, a, t]})        =
    let val id = A.getIdIdent i
	val ar = getArityArgs a
	val _  = addOneId id
	val _  = addArity id id ar
    in []
    end
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_OCONS),   label, value, regions, parents, children = [i, a, t, v, s]})  =
    let val id = A.getIdIdent i
	val ar = getArityArgs a
	val _  = addOneId id
	val _  = addArity id id ar
    in []
    end
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_PARAM), label, value, regions, parents, children = [i, _]})           =
    let val id = A.getIdIdent i
	val nb = getNbOfId id
	val _  =
	    case EN.getScheme id (nb + 1) (getUnifEnv ()) of
		NONE => raise EH.Impossible ("Couldn't find " ^ id ^ "'s environment")
	      | SOME scheme => addParam id scheme
	val _  = addOneId id
	val _  = addArity id id []
    in []
    end
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_TYPARAM), label, value, regions, parents, children = [s, tc, _]})      =
    let val t1  = (*E.mk_type_typ1*) E.mk_type_type D.dummy_label
	val typ = foldr (fn (_, term) => E.mk_type_arrow (t1, term) D.dummy_label)
			t1
			(List.tabulate (getArityTyvarSeq s, fn _ => ()))
	val _   = addParam (A.getIdIdent tc) (E.mk_new_scheme [] typ)
    in []
    end
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_ETPARAM), label, value, regions, parents, children = [s, tc, i, _]})   =
    let val tc  = A.getIdIdent tc
	val ar  = getArityTyvarSeq s
	val eq  = A.getIdIdent i
	val nb  = getNbOfId eq
	val _   =
	    case EN.getScheme eq (nb + 1) (getUnifEnv ()) of
		NONE => raise EH.Impossible ("Couldn't find " ^ eq ^ "'s environment")
	      | SOME scheme => addParam eq scheme
	val t1  = (*E.mk_type_typ1*) E.mk_type_type D.dummy_label
	val typ = foldr (fn (_, term) => E.mk_type_arrow (t1, term) D.dummy_label)
			t1
			(List.tabulate (ar, fn _ => ()))
	val _   = addParam tc (E.mk_new_scheme [] typ)
	val _   = addOneId eq
	val _   = addEqDec tc eq
	val _   = addArity eq eq []
    in []
    end
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_TYCON),   label, value, regions, parents, children = [s, tc, ty]})      =
    let val _ = rmEqDec (A.getIdIdent tc)
    in []
    end
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_EQTYCON), label, value, regions, parents, children = [s, tc, i, ty]})   =
    let val tc = A.getIdIdent tc
	val eq = A.getIdIdent i
	val _  = addOneId eq
	val _  = addEqDec tc eq
	val _  = addArity eq eq []
    in []
    end
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_TYFUN),   label, value, regions, parents, children = [seq, tc, ty]})    = []
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_MSGS),    label, value, regions, parents, children})                    = List.concat (map toNuprlTAtom children)
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_PARSE),   label, value, regions, parents, children = []})               = raise EH.Impossible "term is not parsable"
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_TYPEOF),  label, value, regions, parents, children = [id]})             = []
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_MAIN),    label, value, regions, parents, children = [exp]})            =
    let val nt   = toNuprlTerm exp
	val name = base ^ "_main"
	val prod = NT.mk_nuprl_prod_term NT.mk_nuprl_loc_term (NT.mk_nuprl_msg_term "i")
	val cls  = NT.mk_nuprl_class_term prod
	val res  = genAbstractionAndWF 9 value name nt exp (SOME (cls, [])) true NONE
	val _    = resetDecFP ()
    in res
    end
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_SPEC),    label, value, regions, parents, children = [id1, id2]})       = []
  | toNuprlDec base (term as A.N {kind = (A.DEC, A.DEC_IMPORT),  label, value, regions, parents, children})                    = []
  | toNuprlDec base term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

fun generateHeaders base =
    let val hdrs  = getHeaders ()
    in if not (null hdrs)
       then let val name_all      = base ^ "_headers"
		val name_noinputs = base ^ "_headers_no_inputs"
		val n_id_all      = mk_nuprl_user_term name_all      []
		val n_id_noinputs = mk_nuprl_user_term name_noinputs []
		val (n_lst_all, n_lst_noinputs) =
		    foldr (fn ((kind, hdr), (lst_all, lst_noinputs)) =>
			      let val lst_all' = NT.mk_nuprl_cons_term hdr lst_all
				  val lst_noinputs' =
				      if kind = "input"
				      then lst_noinputs
				      else NT.mk_nuprl_cons_term hdr lst_noinputs
			      in (lst_all', lst_noinputs')
			      end)
			  (NT.mk_nuprl_nil_term, NT.mk_nuprl_nil_term)
			  hdrs
		val n_typ = NT.mk_nuprl_list_term NT.mk_nuprl_name_term
		fun gen name id lst =
		    let val n_abs = NT.mk_nuprl_iabstraction_term id lst
			val n_mem = NT.mk_nuprl_member_term n_typ id
			val n_thm = NT.mk_nuprl_itheorem_term (get_user_name name ^ "_wf") n_mem
		    in [n_abs, n_thm]
		    end
		val objs1 = gen name_all      n_id_all      n_lst_all
		val objs2 = gen name_noinputs n_id_noinputs n_lst_noinputs
		val _     = resetHeaders ()
	    in SOME (n_id_noinputs, objs1 @ objs2)
	    end
       else NONE
    end

fun getHeadersObjs (SOME (_, objs)) = objs
  | getHeadersObjs NONE = []

fun getHeadersNoInputs (SOME (id, _)) = id
  | getHeadersNoInputs NONE = NT.mk_nuprl_nil_term

fun generateStdMa base id term =
    (case A.getKind term of
	 (A.DEC, A.DEC_MAIN) =>
	 (case A.getChildren term of
	      [exp] =>
	      let val name_stdma      = base ^ "_stdma"
		  val name_constraint = base ^ "_message-constraint"
		  val name_delivered  = base ^ "_messages-delivered"
		  val nt              = toNuprlTerm exp
		  val set             = get_all_ids_list [nt]
		  val eo              = A.newIdIdSet set "eo"
		  val neo             = NT.mk_nuprl_variable_term eo
		  val teo             = NT.mk_nuprl_event_ordering_p_term
		  val propip          = NT.mk_nuprl_prop_term "i'"
		  val nity            = NT.mk_nuprl_fun_term teo propip
		  val lvl             = SOME "i"
		  val stdma           = NT.mk_nuprl_std_ma_term             neo nt id
		  val constraint      = NT.mk_nuprl_message_constraint_term neo nt id
		  val delivered       = NT.mk_nuprl_messages_delivered_term neo nt
		  val lstdma          = NT.mk_nuprl_lambda_term eo stdma
		  val lconstraint     = NT.mk_nuprl_lambda_term eo constraint
		  val ldelivered      = NT.mk_nuprl_lambda_term eo delivered
		  fun toAbs n name nterm =
		      genAbstractionAndWF n "" name nterm exp (SOME (nity, [])) false lvl
		  val res =
		      toAbs 10 name_stdma      lstdma      @
		      toAbs 11 name_constraint lconstraint @
		      toAbs 12 name_delivered  ldelivered
		  val _ = resetDecFP ()
	      in res
	      end
	    | _ => [])
       | (A.DEC, _) => []
       | _ => List.concat (map (generateStdMa base id) (A.getChildren term)))

fun getBasePrefix base [] = (base, "", [])
  | getBasePrefix base (lst as (spec :: terms)) =
    if A.getKind spec = (A.DEC, A.DEC_SPEC)
    then case A.getChildren spec of
	     [id1, id2] =>
	     let val name = A.getIdIdent id1
		 val pref = A.getIdIdent id2
	     in (name, pref, terms)
	     end
	   | _ => raise EH.Impossible A.wrongFormat
    else (base, "", lst)

(* FILE *)
fun toNuprlFile fbase (term as A.N {kind = (A.FILE, A.FILE_F), label, value, regions, parents, children}) =
    let val (base, pref, children') = getBasePrefix fbase children
	val _     = setPrefix pref
	val terms = List.concat (map (fn term => toNuprlDec base term) children')
	val hop   = generateHeaders base
	val hdrs  = getHeadersObjs hop
	val stdma = generateStdMa base (getHeadersNoInputs hop) term
	val _     = resetPrefix ()
    in terms @ hdrs @ stdma
    end
  | toNuprlFile fbase term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

(* PROG *)
fun toNuprlProg fbase (term as A.N {kind = (A.PROG, A.PROG_P), label, value, regions, parents, children}) =
    List.concat (map (fn term => toNuprlFile fbase term) children)
  | toNuprlProg fbase term = (print (A.toString term); raise EH.Impossible A.wrongFormat)

fun generateProperties fname =
    let val st1   = NT.mk_nuprl_istring_term fname
	val prop1 = NT.mk_nuprl_iproperty_term fname st1
	val cons1 = NT.mk_nuprl_icons_cons_term prop1 NT.mk_nuprl_icons_nil_term
    in cons1
    end

fun loadNuprlPrelude () =
    let val lst = [("inl",  "inl",  [("x", [])]),
		   ("inr",  "inr",  [("x", [])]),
		   ("isl",  "isl",  [("x", [])]),
		   ("!",    "bnot", [("x", [])]),
		   ("fst",  "pi1",  [("x", [])]),
		   ("snd",  "pi2",  [("x", [])]),
		   ("eqof", "eqof", [("x", [])])]
    in app (fn (id, name, ar) => (addArity id name ar; ())) lst
    end

fun toNuprl file obid term env sub =
    let val _       = A.setIdsTerm term
	val _       = reset ()
	val _       = setObid obid
	val _       = loadNuprlPrelude ()
	val solved  = EN.solver env sub
	val _       = setUnifEnv solved
	val fname   = #file (OS.Path.splitDirFile file)
	val fbase   = #base (OS.Path.splitBaseExt fname)
	val terms   = toNuprlProg fbase term
	val cons1   = generateProperties fname
	val n_terms = (*cons1 ::*) terms
	val _       = A.resetSimp ()
    in n_terms
    end


(* ------ EXPORT NUPRL TO E# ------ *)

(*fun prelude () =
    let val list_ind =
	    let val nilcase  = A.idToExp "nilcase"
		val patnil   = A.idToPat "nil"
		val branch1  = A.mk_new_term A.MRULE_M "" [] [patnil, nilcase]
		val conscase = A.idToExp "conscase"
		val x_id     = A.mk_new_term A.ID_VID "x"  [] []
		val xs_id    = A.mk_new_term A.ID_VID "xs" [] []
		val z_id     = A.mk_new_term A.ID_VID "z"  [] []
		val x_atexp  = A.mk_new_term A.ATEXP_ID "" [] [x_id]
		val xs_atexp = A.mk_new_term A.ATEXP_ID "" [] [xs_id]
		val z_atexp  = A.mk_new_term A.ATEXP_ID "" [] [z_id]
		val app1     = A.mk_new_term A.EXP_APP "" [] [conscase, x_atexp]
		val app2     = A.mk_new_term A.EXP_APP "" [] [app1, xs_atexp]
		val app3     = A.mk_new_term A.EXP_APP "" [] [app2, z_atexp]
		val lam      = A.mk_new_term A.EXP_LAMBDA "" [] [z_id, app3]
		val patcons  = A.mk_new_term A.PAT_CONS "." [] [A.idToPat "x", A.idToPat "xs"]
		val branch2  = A.mk_new_term A.MRULE_M "" [] [patcons, lam]
		val atwild   = A.mk_new_term A.ATPAT_WILD "" [] []
		val patwild  = A.mk_new_term A.PAT_ATPAT "" [] [atwild]
		(* NOTE: for the last branch, we don't case what's the expression,
		 * this branch is never gonna be reached. *)
		val branch3  = A.mk_new_term A.MRULE_M "" [] [patwild, nilcase]
		val match    = A.mk_new_term A.MATCH_M "" [] [branch1, branch2]
		val m_exp    = A.mk_new_term A.EXP_MATCH "" [] [A.idToExp "L", match]
		val L_id     = A.mk_new_term A.ID_VID "L" [] []
		val nil_id   = A.mk_new_term A.ID_VID "nilcase" [] []
		val cons_id  = A.mk_new_term A.ID_VID "conscase" [] []
		val L_arg    = A.mk_new_term A.ATPAT_ID "" [] [L_id]
		val nil_arg  = A.mk_new_term A.ATPAT_ID "" [] [nil_id]
		val cons_arg = A.mk_new_term A.ATPAT_ID "" [] [cons_id]
		val args     = A.mk_new_term A.PARAM_P "" [] [L_arg, nil_arg, cons_arg]
		val li_id    = A.mk_new_term A.ID_VID "list_ind" [] []
		val bind     = A.mk_new_term A.BIND_DEC "" [] [li_id, args, m_exp]
	    in A.mk_new_term A.DEC_LET "" [] [bind]
	    end
    in [list_ind]
    end*)

fun fromNuprlExp (NT.TERM ((("lambda", "OPID"), parameters), [NT.B_TERM ([var], term)])) =
    let val id  = A.mk_new_term A.ID_VID var [] []
	val exp = fromNuprlExp term
    in A.mk_new_term A.EXP_LAMBDA "" [] [id, exp]
    end
  | fromNuprlExp (NT.TERM ((("natural_number", "OPID"), [(n, "n")]), [])) =
    (case Int.fromString n of
	 NONE     => raise EH.Impossible "wrong format"
       | SOME nat =>
	 let val sc    = A.mk_new_term A.SCON_INT n [] []
	     val atexp = A.mk_new_term A.ATEXP_SCON "" [] [sc]
	 in A.mk_new_term A.EXP_ATEXP "" [] [atexp]
	 end)
  | fromNuprlExp (NT.TERM ((("variable", "OPID"), [(var, "v")]), [])) = A.idToExp var
  | fromNuprlExp (NT.TERM ((("apply", "OPID"), []), [NT.B_TERM ([], term1), NT.B_TERM ([], term2)])) =
    let val exp1  = fromNuprlExp term1
	val exp2  = fromNuprlExp term2
	val atexp = A.mk_new_term A.ATEXP_PAREN "" [] [exp2]
    in A.mk_new_term A.EXP_APP "" [] [exp1, atexp]
    end
  | fromNuprlExp (NT.TERM ((("list_ind", "OPID"), []), [NT.B_TERM ([], L), NT.B_TERM ([], nilcase), NT.B_TERM ([x, xs, r], conscase)])) =
    let val li_exp     = A.idToExp "list_ind"
	val L_exp      = fromNuprlExp L
	val L_atexp    = A.mk_new_term A.ATEXP_PAREN "" [] [L_exp]
	val nil_exp    = fromNuprlExp nilcase
	val nil_atexp  = A.mk_new_term A.ATEXP_PAREN "" [] [nil_exp]
	val x_id       = A.mk_new_term A.ID_VID x  [] []
	val xs_id      = A.mk_new_term A.ID_VID xs [] []
	val r_id       = A.mk_new_term A.ID_VID r  [] []
	val cons_exp   = fromNuprlExp conscase
	val lam1       = A.mk_new_term A.EXP_LAMBDA "" [] [r_id,  cons_exp]
	val lam2       = A.mk_new_term A.EXP_LAMBDA "" [] [xs_id, lam1]
	val lam3       = A.mk_new_term A.EXP_LAMBDA "" [] [x_id,  lam2]
	val lam_atexp  = A.mk_new_term A.ATEXP_PAREN "" [] [lam3]
	val app1       = A.mk_new_term A.EXP_APP "" [] [li_exp, L_atexp]
	val app2       = A.mk_new_term A.EXP_APP "" [] [app1, nil_atexp]
	val app3       = A.mk_new_term A.EXP_APP "" [] [app2, lam_atexp]
    in app3
    end
  (* TODO: Othersize we've got to look it up *)
  | fromNuprlExp _ = raise EH.Impossible "can't import the Nuprl term"

fun fromNuprlLhsRhs id lhs rhs =
    case lhs of
	NT.TERM (((opid, "OPID"), parameters), []) =>
	if opid = id
	then let val id   = A.mk_new_term A.ID_VID id [] []
		 val exp  = fromNuprlExp rhs
		 val list = map (fn (value, kind) =>
				    let val id    = A.mk_new_term A.ID_VID value [] []
				    in A.mk_new_term A.ATPAT_ID "" [] [id]
				    end)
				  parameters
		 val args = A.mk_new_term A.PARAM_P  "" [] list
		 val bind = A.mk_new_term A.BIND_DEC "" [] [id, args, exp]
		 val dec  = A.mk_new_term A.DEC_LET  "" [] [bind]
	     in dec
	     end
	else raise EH.Impossible "wrong term"
      | _ => raise EH.Impossible "wrong format"

end
