(* Copyright 2011 Cornell University
 *
 *
 * This file is part of EventML - a tool aiming at specifying
 * distributed protocols in an ML like language.  It is an interface
 * to the logic of events and is compiled into Nuprl.  It is written
 * by the NUPRL group of Cornell University, Ithaca, NY.
 *
 * EventML is a free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EventML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EventML.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  o Authors:     Vincent Rahli
 *  o Affiliation: Cornell University, NUPRL group
 *  o Date:        29 July 2011
 *  o File name:   NuprlTerms.sml
 *  o Description: Datatype for nuprl terms
 *)


structure NuprlTerms :> NUPRL_TERMS = struct

structure T  = ListFormat
structure EH = LibBase

type variable             = string

type opid                 = string
type tag                  = string

type parameter_kind       = string (* such as: "token", "nat", "level-expression" *)
type parameter_value      = string

type opid_tag             = opid * tag
type parameter            = parameter_value * parameter_kind

type operator             = opid_tag * parameter list

datatype nuprl_term       = TERM   of operator * nuprl_bound_term list
     and nuprl_bound_term = B_TERM of variable list * nuprl_term


(* ------ TO STRING ------ *)

fun toStringOpid  opid  = opid
fun toStringTag   tag   = tag
fun toStringValue value = value
fun toStringKind  kind  = kind

fun toStringParameter (value, kind) = toStringValue value ^ ":" ^ toStringKind kind

fun toStringParameters params =
    T.fmt {init  = "",
	   final = "",
	   sep   = ",",
	   fmt   = toStringParameter}
	  params

fun toStringOpidTag (opid, tag) = toStringOpid opid ^ ":" ^ toStringTag tag

fun toStringOperator (opid_tag, []) = toStringOpidTag opid_tag
  | toStringOperator (opid_tag, parameters) =
    toStringOpidTag opid_tag ^ "," ^ toStringParameters parameters

fun toStringVars vars =
    T.fmt {init  = "",
	   final = "",
	   sep   = ",",
	   fmt   = fn v => v ^ ":v"}
	  vars

fun toStringTerm (TERM (operator, [])) ind =
    "{" ^ toStringOperator operator   ^ "}()"
  | toStringTerm (TERM (operator, bterms)) ind =
    let val ind' = ind ^ " "
    in "{" ^ toStringOperator operator    ^ "}\n" ^ ind ^
       "(" ^ toStringBTerms   bterms ind' ^ ")"
    end

and toStringBTerm (B_TERM ([],   term)) ind = toStringTerm term ind
  | toStringBTerm (B_TERM (vars, term)) ind =
    "{bound_id:OPID," ^ toStringVars vars ^ "}\n" ^ ind ^
    "(" ^ toStringTerm term (ind ^ " ") ^ ")"

and toStringBTerms bterms ind =
    T.fmt {init  = "",
	   final = "",
	   sep   = ";\n" ^ ind,
	   fmt   = fn bterm => toStringBTerm bterm ind}
	  bterms

and toStringTerms terms ind =
    T.fmt {init  = "",
	   final = "",
	   sep   = ";\n" ^ ind,
	   fmt   = fn term => toStringTerm term ind}
	  terms

val toStringTerm = fn term => toStringTerm term ""

fun ppTerm (TERM ((("apply", _), []), [B_TERM ([], f),
				       B_TERM ([], a)])) = "(" ^ ppTerm f ^ ")(" ^ ppTerm a ^ ")"
  | ppTerm (TERM ((("int_eq", _), []), [B_TERM ([], x),
					B_TERM ([], y),
					B_TERM ([], w),
					B_TERM ([], z)])) = "if " ^ ppTerm x ^ " = " ^ ppTerm y ^ " then " ^ ppTerm w ^ " else " ^ ppTerm z
  | ppTerm (TERM ((("subtract", _), []), [B_TERM ([], x),
					  B_TERM ([], y)])) = ppTerm x ^ "-" ^ ppTerm y
  | ppTerm (TERM ((("add", _), []), [B_TERM ([], x),
				     B_TERM ([], y)])) = ppTerm x ^ "+" ^ ppTerm y
  | ppTerm (TERM ((("pair", _), []), [B_TERM ([], x),
				      B_TERM ([], y)])) = "(" ^ ppTerm x ^ "," ^ ppTerm y ^ ")"
  | ppTerm (TERM ((("variable", _),       [(v,vkind)]), [])) = v
  | ppTerm (TERM ((("natural_number", _), [(n,nkind)]), [])) = n
  | ppTerm (TERM ((("ycomb", _), []), [])) = "Y"
  | ppTerm (TERM ((("lambda", _), []), [B_TERM ([x], f)])) = "\\" ^ x ^ ". " ^ ppTerm f
  | ppTerm (TERM ((("inr", _), []), [B_TERM ([], t)])) = "inr(" ^ ppTerm t ^ ")"
  | ppTerm (TERM ((("inl", _), []), [B_TERM ([], t)])) = "inl(" ^ ppTerm t ^ ")"
  | ppTerm (TERM ((("decide", _), []), [B_TERM ([], dec),
					B_TERM ([v1], t1),
					B_TERM ([v2], t2)])) = "case " ^ ppTerm dec ^ " of inl("^ v1 ^ ") => " ^ ppTerm t1 ^ " | inr(" ^ v2 ^ ") => " ^ ppTerm t2
  | ppTerm (TERM ((("spread", _), []), [B_TERM ([], pair),
					B_TERM ([x1,x2], t)])) = "let " ^ x1 ^ "," ^ x2 ^ " = " ^ ppTerm pair ^ " in "^ ppTerm t
  | ppTerm (TERM ((("list_ind", _), []), [B_TERM ([], L),
					  B_TERM ([], base),
					  B_TERM ([x,xs], reccase)])) = "rec-case " ^ ppTerm L ^ " of [] => " ^ ppTerm base ^ " | " ^ x ^ "." ^ xs ^ " => " ^ ppTerm reccase
  | ppTerm (TERM ((("callbyvalueall", _), []), [B_TERM ([], arg),
						B_TERM ([x], B)])) = "let " ^ x ^ " := " ^ ppTerm arg ^ " in " ^ ppTerm B
  | ppTerm term = toStringTerm term


(* ------ ACCESSORS ------ *)

fun opid_of_term (TERM (((opid, _), _), _)) = opid

fun parameters_of_term (TERM (((_, _), params), _)) = params

fun bterms_of_term (TERM (_, bterms)) = map (fn (B_TERM x) => x) bterms

fun subterms term = map (fn (_, term) => term) (bterms_of_term term)

fun subtermn n (TERM (_, bterm_list)) =
    case List.nth (bterm_list, n - 1) of
	B_TERM ([], t) => t
      | _ => raise Fail "subtermn"

fun type_of_parameter  (_, kind)  = kind
fun value_of_parameter (value, _) = value

fun destruct_natural_parameter (n,"n") = Option.valOf (Int.fromString n)
  | destruct_natural_parameter _ = raise Fail "destruct_natural_parameter"

fun firstnat term = destruct_natural_parameter (hd (parameters_of_term term))


(* ------ CHECKERS ------ *)

fun is_nuprl_term token term = (opid_of_term term = token)

fun is_nuprl_natural_number_term term = is_nuprl_term "natural_number" term
fun is_nuprl_minus_term          term = is_nuprl_term "minus"          term
fun is_nuprl_pair_term           term = is_nuprl_term "pair"           term
fun is_nuprl_lambda_term         term = is_nuprl_term "lambda"         term
fun is_nuprl_variable_term       term = is_nuprl_term "variable"       term
fun is_nuprl_type_term           term = is_nuprl_term "universe"       term
fun is_nuprl_function_term       term = is_nuprl_term "function"       term
fun is_nuprl_iwf_lemmas_term     term = is_nuprl_term "!wf"            term
fun is_nuprl_iabstraction_term   term = is_nuprl_term "!abstraction"   term

fun is_nuprl_iwftheorem_term (TERM ((("!theorem", _), [(name, "t")]), [B_TERM ([], theorem)])) =
    String.isSuffix "_wf" name
  | is_nuprl_iwftheorem_term _ = false

fun equal_parameters param1 param2 = (param1 : parameter) = param2

(* ------ DESTRUCTORS ------ *)

fun dest_simple_term term =
    let val opid  = opid_of_term term
	val terms =
	    map (fn ([], term) => term
		  | _ => (print (toStringTerm term); raise EH.Impossible "dest_simple_term"))
		(bterms_of_term term)
    in (opid, terms)
    end

fun dest_simple_null_term term =
    let val opid  = opid_of_term term
	val terms =
	    map (fn ([],   term) => term
		  | ([""], term) => term
		  | _ => (print (toStringTerm term); raise EH.Impossible "dest_simple_term"))
		(bterms_of_term term)
    in (opid, terms)
    end

fun dest_term term =
    ((opid_of_term term, parameters_of_term term),
     bterms_of_term term)

fun dest_lambda term =
    if is_nuprl_lambda_term term
    then case bterms_of_term term of
	     [([x], B)] => (x, B)
	   | _ => raise EH.Impossible "dest_lambda"
    else raise EH.Impossible "dest_lambda"

fun dest_pair term =
    if is_nuprl_pair_term term
    then case bterms_of_term term of
	     [([], T1), ([], T2)] => (T1, T2)
	   | _ => raise EH.Impossible "dest_pair"
    else raise EH.Impossible "dest_pair"

fun dest_lambdas term =
    if is_nuprl_lambda_term term
    then case bterms_of_term term of
	     [([x], B)] =>
	     let val (bounds, body) = dest_lambdas B
	     in (x :: bounds, body)
	     end
	   | _ => raise EH.Impossible "dest_lambda"
    else ([], term)

fun dest_variable term =
    if is_nuprl_variable_term term
    then case term of
	     TERM ((("variable", _), [(var, "v")]), _) => var
	   | _ => raise EH.Impossible "dest_variable"
    else raise EH.Impossible "dest_variable"

fun dest_so_variable term =
    if is_nuprl_variable_term term
    then case term of
	     TERM ((("variable", _), [(var, "v")]), bterms) =>
	     let val lst =
		     map (fn B_TERM (vars, t) =>
			     if is_nuprl_variable_term t andalso List.null vars
			     then dest_variable t
			     else raise EH.Impossible "dest_so_variable")
			 bterms
	     in (var, lst)
	     end
	   | _ => raise EH.Impossible "dest_so_variable"
    else raise EH.Impossible "dest_so_variable"

fun dest_eclass (TERM ((("eclass", tag), params),
		       [B_TERM ([], msg), B_TERM ([es, e], X)])) =
    (msg, es, e, X)
  | dest_eclass _ = raise EH.Impossible "dest_eclass"

fun dest_function (TERM ((("function", tag), params),
			 [B_TERM ([], term1), B_TERM ([""], term2)])) =
    (term1, term2)
  | dest_function _ = raise EH.Impossible "dest_function"

fun dest_product (TERM ((("product", tag), params),
			[B_TERM ([], term1), B_TERM ([""], term2)])) =
    (term1, term2)
  | dest_product _ = raise EH.Impossible "dest_product"

fun dest_iabstraction (TERM ((("!abstraction", tag), []), [B_TERM ([], t1), B_TERM ([], t2), B_TERM ([], t3)])) = (t1, t2, t3)
  | dest_iabstraction _ = raise EH.Impossible "dest_iabstraction"

fun dest_member (TERM ((("member", tag), []), [B_TERM ([], typ), B_TERM ([], term)])) = (typ, term)
  | dest_member _ = raise EH.Impossible "dest_member"

fun dest_iwftheorem (TERM ((("!theorem", _), [(name, "t")]), [B_TERM ([], theorem)])) = theorem
  | dest_iwftheorem _ = raise EH.Impossible "dest_iwftheorem"

fun dest_primed_class (TERM ((("primed-class", tag), params), [B_TERM ([], class)])) = class
  | dest_primed_class _ = raise EH.Impossible "dest_primed_class"

fun dest_natural_number (TERM ((("natural_number", tag), [(n, "n")]), [])) = n
  | dest_natural_number (TERM ((("natural_number", tag), [(n, "nat")]), [])) = n
  | dest_natural_number _ = raise EH.Impossible "dest_natural_number"

fun dest_minus (TERM ((("minus", tag), []), [B_TERM ([], t)])) = t
  | dest_minus _ = raise EH.Impossible "dest_minus"

fun dest_integer term =
    let val n =
	    if is_nuprl_natural_number_term term
	    then dest_natural_number term
	    else if is_nuprl_minus_term term
	    then dest_natural_number (dest_minus term)
	    else raise Fail "error format"
    in case Int.fromString n of
	   NONE => raise Fail "error format"
	 | SOME x => x
    end

fun is_nuprl_minus_natural_number_term term =
    is_nuprl_minus_term term
    andalso
    is_nuprl_natural_number_term (dest_minus term)

fun is_nuprl_integer_term term =
    is_nuprl_natural_number_term term
    orelse
    is_nuprl_minus_natural_number_term term


(* ------ CONSTRUCTORS ------ *)

fun mk_nuprl_variable_parameter  tok    = (tok,                "v")
fun mk_nuprl_natural_parameter   tag    = (Int.toString tag,   "n")
fun mk_nuprl_token_parameter     token  = (token,              "t")
fun mk_nuprl_string_parameter    string = (string,             "s")
fun mk_nuprl_level_exp_parameter level  = (level,              "l")
fun mk_nuprl_obid_parameter      obid   = (obid,               "o")
fun mk_nuprl_bool_parameter      bool   = (if bool then "T" else "F", "b")

fun mk_nuprl_parameter (param : parameter) = param

fun mk_nuprl_term (opid, params) b_terms =
    TERM (((opid, "OPID"), params), map (fn x => B_TERM x) b_terms)

fun mk_nuprl_simple_term opid term_list =
    TERM (((opid, "OPID"), []), map (fn term => B_TERM ([], term)) term_list)

fun mk_nuprl_natural_number_term nat =
    TERM ((("natural_number", "OPID"), [mk_nuprl_natural_parameter nat]), [])

fun mk_nuprl_token_term tok =
    TERM ((("token", "OPID"), [mk_nuprl_token_parameter tok]), [])

fun mk_nuprl_minus_term term = mk_nuprl_simple_term "minus" [term]

val mk_nuprl_axiom_term = mk_nuprl_simple_term "axiom" []

fun mk_nuprl_integer_term int =
    if int < 0
    then mk_nuprl_minus_term (mk_nuprl_natural_number_term (~int))
    else mk_nuprl_natural_number_term int

fun mk_nuprl_apply_term func arg = mk_nuprl_simple_term "apply" [func, arg]

fun mk_nuprl_applies_term func args =
    foldl (fn (arg, f) => mk_nuprl_simple_term "apply" [f, arg])
	  func
	  args

fun mk_nuprl_let_term var pat body =
    mk_nuprl_term ("let", []) [([], pat), ([var], body)]

fun mk_nuprl_pair_term left right = mk_nuprl_simple_term "pair" [left, right]

fun mk_nuprl_variable_term tok = mk_nuprl_term ("variable", [mk_nuprl_variable_parameter tok]) []

fun mk_nuprl_lambda_term var term = mk_nuprl_term ("lambda", []) [([var], term)]

fun mk_nuprl_lambdas_term vars term =
    foldr (fn (var, term) => mk_nuprl_lambda_term var term)
	  term
	  vars

fun mk_nuprl_spread_term  pair (v1, v2, bterm) = mk_nuprl_term ("spread",  []) [([], pair), ([v1, v2], bterm)]

fun mk_nuprl_spreadn_term pair ([],       bterm) = raise EH.Impossible ""
  | mk_nuprl_spreadn_term pair ([_],      bterm) = raise EH.Impossible ""
  | mk_nuprl_spreadn_term pair ([v1, v2], bterm) = mk_nuprl_spread_term pair (v1, v2, bterm)
  | mk_nuprl_spreadn_term pair (vars,     bterm) = mk_nuprl_term ("spreadn", []) [([], pair), (vars,     bterm)]

fun mk_nuprl_inl_term term = mk_nuprl_simple_term "inl" [term]
fun mk_nuprl_inr_term term = mk_nuprl_simple_term "inr" [term]

fun mk_nuprl_decide_term dec (var1, bterm1) (var2, bterm2) =
    mk_nuprl_term ("decide", []) [([], dec), ([var1], bterm1), ([var2], bterm2)]

val mk_nuprl_ycomb_term = mk_nuprl_simple_term "ycomb" []

fun mk_nuprl_rec_ind_term arg (f, x, B) =
    mk_nuprl_term ("rec_ind", []) [([], arg), ([f, x], B)]

fun mk_nuprl_genrec_term (n, r, B) =
    mk_nuprl_term ("genrec", []) [([n, r], B)]

fun mk_nuprl_bind_class_term f (x, g) =
    mk_nuprl_term ("bind-class", []) [([], f), ([x], g)]

fun mk_nuprl_combined_class_term         f lst = mk_nuprl_simple_term "simple-comb"            [f, lst]
fun mk_nuprl_rec_combined_class_term     f lst = mk_nuprl_simple_term "rec-combined-class"     [lst, f]
fun mk_nuprl_combined_loc_class_term     f lst = mk_nuprl_simple_term "simple-loc-comb"        [f, lst]
fun mk_nuprl_rec_combined_loc_class_term f lst = mk_nuprl_simple_term "rec-combined-loc-class" [lst, f]

fun mk_nuprl_rec_comb_term f classes init = mk_nuprl_simple_term "rec-comb" [classes, f, init]

fun mk_nuprl_so_apply1_term f x     = mk_nuprl_simple_term "so_apply" [f, x]
fun mk_nuprl_so_apply2_term f x y   = mk_nuprl_simple_term "so_apply" [f, x, y]
fun mk_nuprl_so_apply3_term f x y z = mk_nuprl_simple_term "so_apply" [f, x, y, z]

fun mk_nuprl_combined0_class_term bag =
    mk_nuprl_simple_term "simple-comb-0" [bag]
fun mk_nuprl_combined1_class_term f class =
    mk_nuprl_simple_term "simple-comb-1" [f, class]
fun mk_nuprl_combined2_class_term f class1 class2 =
    mk_nuprl_simple_term "simple-comb-2" [f, class1, class2]
fun mk_nuprl_combined3_class_term f class1 class2 class3 =
    mk_nuprl_simple_term "simple-comb-3" [f, class1, class2, class3]

fun mk_nuprl_combined0_loc_class_term bag =
    mk_nuprl_simple_term "simple-loc-comb-0" [bag]
fun mk_nuprl_combined1_loc_class_term f class =
    mk_nuprl_simple_term "simple-loc-comb-1" [f, class]
fun mk_nuprl_combined2_loc_class_term f class1 class2 =
    mk_nuprl_simple_term "simple-loc-comb-2" [f, class1, class2]
fun mk_nuprl_combined3_loc_class_term f class1 class2 class3 =
    mk_nuprl_simple_term "simple-loc-comb-3" [f, class1, class2, class3]

fun mk_nuprl_rec_combined0_class_term bag =
    mk_nuprl_simple_term "rec-combined-class-0" [bag]
fun mk_nuprl_rec_combined1_class_term f class =
    mk_nuprl_simple_term "rec-combined-class-1" [f, class]
fun mk_nuprl_rec_combined2_class_term f class1 class2 =
    mk_nuprl_simple_term "rec-combined-class-2" [f, class1, class2]
fun mk_nuprl_rec_combined3_class_term f class1 class2 class3 =
    mk_nuprl_simple_term "rec-combined-class-3" [f, class1, class2, class3]

fun mk_nuprl_rec_combined0_class_opt_term opt bag =
    mk_nuprl_simple_term "rec-combined-class-opt-0" [bag, opt]
fun mk_nuprl_rec_combined1_class_opt_term opt f class =
    mk_nuprl_simple_term "rec-combined-class-opt-1" [f, opt, class]
fun mk_nuprl_rec_combined2_class_opt_term opt f class1 class2 =
    mk_nuprl_simple_term "rec-combined-class-opt-2" [f, opt, class1, class2]
fun mk_nuprl_rec_combined3_class_opt_term opt f class1 class2 class3 =
    mk_nuprl_simple_term "rec-combined-class-opt-3" [f, opt, class1, class2, class3]

fun mk_nuprl_rec_combined0_loc_class_term bag =
    mk_nuprl_simple_term "rec-combined-loc-class-0" [bag]
fun mk_nuprl_rec_combined1_loc_class_term f class =
    mk_nuprl_simple_term "rec-combined-loc-class-1" [f, class]
fun mk_nuprl_rec_combined2_loc_class_term f class1 class2 =
    mk_nuprl_simple_term "rec-combined-loc-class-2" [f, class1, class2]
fun mk_nuprl_rec_combined3_loc_class_term f class1 class2 class3 =
    mk_nuprl_simple_term "rec-combined-loc-class-3" [f, class1, class2, class3]

fun mk_nuprl_rec_combined0_loc_class_opt_term opt bag =
    mk_nuprl_simple_term "rec-combined-loc-class-opt-0" [bag, opt]
fun mk_nuprl_rec_combined1_loc_class_opt_term opt f class =
    mk_nuprl_simple_term "rec-combined-loc-class-opt-1" [f, opt, class]
fun mk_nuprl_rec_combined2_loc_class_opt_term opt f class1 class2 =
    mk_nuprl_simple_term "rec-combined-loc-class-opt-2" [f, opt, class1, class2]
fun mk_nuprl_rec_combined3_loc_class_opt_term opt f class1 class2 class3 =
    mk_nuprl_simple_term "rec-combined-loc-class-opt-3" [f, opt, class1, class2, class3]

fun mk_nuprl_lifting0_term f = mk_nuprl_simple_term "lifting-0" [f]
fun mk_nuprl_lifting1_term f = mk_nuprl_simple_term "lifting-1" [f]
fun mk_nuprl_lifting2_term f = mk_nuprl_simple_term "lifting-2" [f]
fun mk_nuprl_lifting3_term f = mk_nuprl_simple_term "lifting-3" [f]
fun mk_nuprl_lifting4_term f = mk_nuprl_simple_term "lifting-4" [f]

fun mk_nuprl_lifting_gen_term n f = mk_nuprl_simple_term "lifting-gen" [n, f]

fun mk_nuprl_lifting_loc0_term f = mk_nuprl_simple_term "lifting-loc-0" [f]
fun mk_nuprl_lifting_loc1_term f = mk_nuprl_simple_term "lifting-loc-1" [f]
fun mk_nuprl_lifting_loc2_term f = mk_nuprl_simple_term "lifting-loc-2" [f]
fun mk_nuprl_lifting_loc3_term f = mk_nuprl_simple_term "lifting-loc-3" [f]
fun mk_nuprl_lifting_loc4_term f = mk_nuprl_simple_term "lifting-loc-4" [f]

fun mk_nuprl_lifting_loc_gen_term n f = mk_nuprl_simple_term "lifting-loc-gen" [n, f]

fun mk_nuprl_concat_lifting0_term f = mk_nuprl_simple_term "concat-lifting-0" [f]
fun mk_nuprl_concat_lifting1_term f = mk_nuprl_simple_term "concat-lifting-1" [f]
fun mk_nuprl_concat_lifting2_term f = mk_nuprl_simple_term "concat-lifting-2" [f]
fun mk_nuprl_concat_lifting3_term f = mk_nuprl_simple_term "concat-lifting-3" [f]
fun mk_nuprl_concat_lifting4_term f = mk_nuprl_simple_term "concat-lifting-4" [f]

fun mk_nuprl_concat_lifting_gen_term n f = mk_nuprl_simple_term "concat-lifting-gen" [n, f]

fun mk_nuprl_concat_lifting_loc0_term f = mk_nuprl_simple_term "concat-lifting-loc-0" [f]
fun mk_nuprl_concat_lifting_loc1_term f = mk_nuprl_simple_term "concat-lifting-loc-1" [f]
fun mk_nuprl_concat_lifting_loc2_term f = mk_nuprl_simple_term "concat-lifting-loc-2" [f]
fun mk_nuprl_concat_lifting_loc3_term f = mk_nuprl_simple_term "concat-lifting-loc-3" [f]
fun mk_nuprl_concat_lifting_loc4_term f = mk_nuprl_simple_term "concat-lifting-loc-4" [f]

fun mk_nuprl_concat_lifting_loc_gen_term n f = mk_nuprl_simple_term "concat-lifting-loc-gen" [n, f]

fun mk_nuprl_make_msg_term hdr typ v = mk_nuprl_simple_term "make-Msg" [hdr, typ, v]

fun mk_nuprl_add_term       term1 term2 = mk_nuprl_simple_term "add"       [term1, term2]
fun mk_nuprl_subtract_term  term1 term2 = mk_nuprl_simple_term "subtract"  [term1, term2]
fun mk_nuprl_multiply_term  term1 term2 = mk_nuprl_simple_term "multiply"  [term1, term2]
fun mk_nuprl_remainder_term term1 term2 = mk_nuprl_simple_term "remainder" [term1, term2]
fun mk_nuprl_divide_term    term1 term2 = mk_nuprl_simple_term "divide"    [term1, term2]

fun mk_nuprl_lt_int_term    term1 term2 = mk_nuprl_simple_term "lt_int" [term1, term2]

fun mk_nuprl_fun_term   term1 term2 =
    mk_nuprl_term ("function", []) [([], term1), ([""], term2)]

fun mk_nuprl_prod_term  term1 term2 =
    mk_nuprl_term ("product", []) [([], term1), ([""], term2)]

fun mk_nuprl_union_term term1 term2 =
    mk_nuprl_simple_term "union" [term1, term2]

val mk_nuprl_int_term  = mk_nuprl_simple_term "int"  []
val mk_nuprl_real_term = mk_nuprl_simple_term "real" []
val mk_nuprl_bool_term = mk_nuprl_simple_term "bool" []
val mk_nuprl_unit_term = mk_nuprl_simple_term "unit" []
val mk_nuprl_atom_term = mk_nuprl_simple_term "atom" []
val mk_nuprl_nat_term  = mk_nuprl_simple_term "nat"  []
val mk_nuprl_loc_term  = mk_nuprl_simple_term "Id"   []
val mk_nuprl_name_term = mk_nuprl_simple_term "name" [] (* Atom List *)

fun mk_nuprl_type_term i = mk_nuprl_term ("universe", [mk_nuprl_level_exp_parameter i]) []
fun mk_nuprl_msg_term  i = mk_nuprl_term ("Message",  [mk_nuprl_level_exp_parameter i]) []

fun mk_nuprl_valuealltype_term i = mk_nuprl_term ("vatype", [mk_nuprl_level_exp_parameter i]) []

fun mk_nuprl_list_term  term = mk_nuprl_simple_term "list" [term]
fun mk_nuprl_bag_term   term = mk_nuprl_simple_term "bag"  [term]
fun mk_nuprl_deq_term   term = mk_nuprl_simple_term "deq"  [term]

fun mk_nuprl_class_term term =
    mk_nuprl_term ("eclass", [mk_nuprl_level_exp_parameter "i'"])
		  [([], mk_nuprl_msg_term "i"),
		   (["e", "es"], term)]

fun mk_nuprl_normal_locally_programmable_term typ class =
    mk_nuprl_term ("normal-locally-programmable", [mk_nuprl_level_exp_parameter "i'"])
		  [([], mk_nuprl_msg_term "i"),
		   ([], typ),
		   ([], class)]

val mk_nuprl_nlp_term = mk_nuprl_normal_locally_programmable_term

fun mk_nuprl_programmable_term typ class =
    mk_nuprl_term ("programmable", [mk_nuprl_level_exp_parameter "i'"])
		  [([], mk_nuprl_msg_term "i"),
		   ([], typ),
		   ([], class)]

fun mk_nuprl_std_ma_term eo class headers =
    mk_nuprl_term ("std-ma", [mk_nuprl_level_exp_parameter "i"])
		  [([], eo),
		   ([], class),
		   ([], headers)]

fun mk_nuprl_message_constraint_term eo class headers =
    mk_nuprl_term ("message-constraint", [mk_nuprl_level_exp_parameter "i"])
		  [([], eo),
		   ([], class),
		   ([], headers)]

fun mk_nuprl_messages_delivered_term eo class =
    mk_nuprl_term ("messages-delivered", [mk_nuprl_level_exp_parameter "i"])
		  [([], eo),
		   ([], class)]

val mk_nuprl_event_ordering_p_term =
    mk_nuprl_term ("event-ordering+", [mk_nuprl_level_exp_parameter "i'"])
		  [([], mk_nuprl_msg_term "i")]

fun mk_nuprl_prop_term i =
    mk_nuprl_term ("prop", [mk_nuprl_level_exp_parameter i]) []

fun mk_nuprl_valueall_type_term typ = mk_nuprl_simple_term "valueall-type" [typ]

fun mk_nuprl_implies_term term1 term2 = mk_nuprl_simple_term "implies" [term1, term2]

fun mk_nuprl_proddeq_term term1 term2 = mk_nuprl_simple_term "proddeq" [term1, term2]
fun mk_nuprl_sumdeq_term  term1 term2 = mk_nuprl_simple_term "sumdeq"  [term1, term2]

fun mk_nuprl_product_deq_term typ1 typ2 deq1 deq2 = mk_nuprl_simple_term "product-deq" [typ1, typ2, deq1, deq2]
fun mk_nuprl_union_deq_term   typ1 typ2 deq1 deq2 = mk_nuprl_simple_term "union-deq"   [typ1, typ2, deq1, deq2]

fun mk_nuprl_list_deq_term term = mk_nuprl_simple_term "list-deq" [term]

val mk_nuprl_bool_deq_term = mk_nuprl_simple_term "bool-deq" []
val mk_nuprl_int_deq_term  = mk_nuprl_simple_term "int-deq"  []
val mk_nuprl_atom_deq_term = mk_nuprl_simple_term "atom-deq" []
val mk_nuprl_nat_deq_term  = mk_nuprl_simple_term "nat-deq"  []
val mk_nuprl_loc_deq_term  = mk_nuprl_simple_term "id-deq"   []
val mk_nuprl_unit_deq_term = mk_nuprl_simple_term "unit-deq" []

fun mk_nuprl_eq_prod_term deq1 deq2 nt1 nt2 =
    let	val bdeq = mk_nuprl_proddeq_term deq1 deq2
	val app1 = mk_nuprl_apply_term bdeq nt1
	val app2 = mk_nuprl_apply_term app1 nt2
    in app2
    end

fun mk_nuprl_eq_union_term deq1 deq2 nt1 nt2 =
    let	val bdeq = mk_nuprl_sumdeq_term deq1 deq2
	val app1 = mk_nuprl_apply_term bdeq nt1
	val app2 = mk_nuprl_apply_term app1 nt2
    in app2
    end

fun mk_nuprl_eq_list_term deq nt1 nt2 =
    let	val bdeq = mk_nuprl_list_deq_term deq
	val app1 = mk_nuprl_apply_term bdeq nt1
	val app2 = mk_nuprl_apply_term app1 nt2
    in app2
    end

fun mk_nuprl_eq_bag_term deq nt1 nt2 = mk_nuprl_simple_term "bag-eq" [deq, nt1, nt2]

fun mk_nuprl_eq_atom_term nt1 nt2 = mk_nuprl_simple_term "eq_atom" [nt1, nt2]
fun mk_nuprl_eq_bool_term nt1 nt2 = mk_nuprl_simple_term "eq_bool" [nt1, nt2]
fun mk_nuprl_eq_int_term  nt1 nt2 = mk_nuprl_simple_term "eq_int"  [nt1, nt2]
fun mk_nuprl_eq_loc_term  nt1 nt2 = mk_nuprl_simple_term "eq_id"   [nt1, nt2]

fun mk_nuprl_all_term term1 (var, term2) =
    mk_nuprl_term ("all", []) [([], term1), ([var], term2)]

fun mk_nuprl_uall_term term1 (var, term2) =
    mk_nuprl_term ("uall", []) [([], term1), ([var], term2)]

fun mk_nuprl_member_term term1 term2 = mk_nuprl_simple_term "member" [term1, term2]

fun mk_nuprl_bor_term  term1 term2 = mk_nuprl_simple_term "bor"  [term1, term2]
fun mk_nuprl_band_term term1 term2 = mk_nuprl_simple_term "band" [term1, term2]

fun mk_nuprl_list_ind_term L nilcase (x, xs, r, conscase) =
    mk_nuprl_term ("list_ind", []) [([], L), ([], nilcase), ([x, xs, r], conscase)]

fun mk_nuprl_ind_term i (x, rd, downcase) basecase (y, ru, upcase) =
    mk_nuprl_term ("ind", []) [([], i), ([x, rd], downcase), ([], basecase), ([y, ru], upcase)]

fun mk_nuprl_cons_term elt list = mk_nuprl_simple_term "cons" [elt, list]

fun mk_nuprl_concat_term list1 list2 = mk_nuprl_simple_term "append" [list1, list2]

fun mk_nuprl_select_term n l = mk_nuprl_simple_term "select" [n, l]

fun mk_nuprl_parallel_class_term class1 class2 = mk_nuprl_simple_term "parallel-class" [class1, class2]

fun mk_nuprl_until_class_term class1 class2 = mk_nuprl_simple_term "until-class" [class1, class2]

fun mk_nuprl_primed_class_opt_term class bag = mk_nuprl_simple_term "primed-class-opt" [class, bag]

fun mk_nuprl_class_opt_term class bag = mk_nuprl_simple_term "class-opt" [class, bag]

fun mk_nuprl_class_opt_class_term class1 class2 = mk_nuprl_simple_term "class-opt-class" [class1, class2]

fun mk_nuprl_once_class_term class = mk_nuprl_simple_term "once-class" [class]

fun mk_nuprl_send_once_class_term class = mk_nuprl_simple_term "send-once-class" [class]

fun mk_nuprl_send_once_loc_class_term class = mk_nuprl_simple_term "send-once-loc-class" [class]

fun mk_nuprl_on_loc_class_term class = mk_nuprl_simple_term "on-loc-class" [class]

fun mk_nuprl_class_at_term class locs = mk_nuprl_simple_term "class-at" [class, locs]

fun mk_nuprl_primed_class_term class = mk_nuprl_simple_term "primed-class" [class]

fun mk_nuprl_base_locs_headers_term term1 term2 =
    mk_nuprl_simple_term "base-locs-headers" [term1, term2]

fun mk_nuprl_general_base_class_term term1 term2 =
    mk_nuprl_simple_term "general-base-class" [term1, term2]

fun mk_nuprl_base_headers_msg_val_term term1 term2 =
    mk_nuprl_simple_term "base-headers-msg-val" [term1, term2]

fun mk_nuprl_base_headers_msg_val_loc_term term1 term2 term3 =
    mk_nuprl_simple_term "base-headers-msg-val-loc" [term1, term2, term3]

val mk_nuprl_condition_cons_term = mk_nuprl_simple_term "!condition_cons" []

fun mk_nuprl_iabstraction_term t1 t2 =
    mk_nuprl_simple_term "!abstraction" [mk_nuprl_condition_cons_term, t1, t2]

fun mk_nuprl_itheorem_term name term =
    mk_nuprl_term ("!theorem", [mk_nuprl_token_parameter name]) [([], term)]

fun mk_nuprl_iproperty_term name value =
    mk_nuprl_term ("!property", [mk_nuprl_token_parameter name]) [([], value)]

fun mk_nuprl_istring_term string =
    mk_nuprl_term ("!string", [mk_nuprl_string_parameter string]) []

fun mk_nuprl_ibool_term bool =
    mk_nuprl_term ("!bool", [mk_nuprl_bool_parameter bool]) []

val mk_nuprl_icons_nil_term = mk_nuprl_simple_term "!cons" []

fun mk_nuprl_icons_cons_term term1 term2 = mk_nuprl_simple_term "!cons" [term1, term2]

fun mk_nuprl_iinclude_properties_term prop term =
    mk_nuprl_simple_term "!include_properties" [prop, term]

fun mk_nuprl_iwf_lemmas_term wf_lemmas = mk_nuprl_simple_term "!wf" wf_lemmas

val mk_nuprl_nil_term    = mk_nuprl_simple_term "nil"    []
val mk_nuprl_btrue_term  = mk_nuprl_simple_term "btrue"  []
val mk_nuprl_bfalse_term = mk_nuprl_simple_term "bfalse" []

val mk_nuprl_empty_bag_term = mk_nuprl_simple_term "empty-bag" []

fun mk_nuprl_single_bag_term elt = mk_nuprl_simple_term "single-bag" [elt]

fun mk_nuprl_cons_bag_term head tail = mk_nuprl_simple_term "cons-bag" [head, tail]

fun mk_nuprl_bnot_term term = mk_nuprl_simple_term "bnot" [term]

fun mk_nuprl_callbyvalue_term arg (x, B) =
    mk_nuprl_term ("callbyvalue", []) [([], arg), ([x], B)]

fun mk_nuprl_callbyvalueall_term arg (x, B) =
    mk_nuprl_term ("callbyvalueall", []) [([], arg), ([x], B)]

fun mk_nuprl_ite_term t1 t2 t3 = mk_nuprl_simple_term "ifthenelse" [t1, t2, t3]

fun mk_nuprl_pi1_term term = mk_nuprl_simple_term "pi1" [term]
fun mk_nuprl_pi2_term term = mk_nuprl_simple_term "pi2" [term]

fun mk_nuprl_hd_term term = mk_nuprl_simple_term "hd" [term]
fun mk_nuprl_tl_term term = mk_nuprl_simple_term "tl" [term]

val mk_nuprl_it_term = mk_nuprl_simple_term "it" []

fun mk_nuprl_equal_term term1 term2 ty = mk_nuprl_simple_term "equal" [ty, term1, term2]

(*fun mk_nuprl_mlnk_term term = mk_nuprl_simple_term "mlnk" [term]
fun mk_nuprl_mtag_term term = mk_nuprl_simple_term "mtag" [term]
fun mk_nuprl_mval_term term = mk_nuprl_simple_term "mval" [term]*)

fun mk_nuprl_msg_header_term term = mk_nuprl_simple_term "msg-header" [term]
fun mk_nuprl_msg_type_term   term = mk_nuprl_simple_term "msg-type"   [term]
fun mk_nuprl_msg_body_term   term = mk_nuprl_simple_term "msg-body"   [term]

fun mk_nuprl_msg_has_type_term term1 term2 = mk_nuprl_simple_term "msg-has-type" [term1, term2]

fun mk_nuprl_name_eq_term term1 term2 = mk_nuprl_simple_term "name_eq" [term1, term2]

fun mk_nuprl_reduce_term term1 term2 term3 =
    mk_nuprl_simple_term "reduce" [term1, term2, term3]

fun mk_nuprl_bag_map_term f bag = mk_nuprl_simple_term "bag-map" [f, bag]

fun mk_nuprl_bag_null_term bag = mk_nuprl_simple_term "bag-null" [bag]

fun mk_nuprl_bag_only_term bag = mk_nuprl_simple_term "bag-only" [bag]

fun mk_nuprl_bag_size_term bag = mk_nuprl_simple_term "bag-size" [bag]

fun mk_nuprl_eq_term_term term1 term2 = mk_nuprl_simple_term "eq_term" [term1, term2]

fun mk_nuprl_evalall_term term = mk_nuprl_simple_term "evalall" [term]

fun build_primitive_value term params =
    let val opid = opid_of_term term
    in if List.exists (fn x => x = opid) ["inl", "inr", "pair", "cons"]
       then mk_nuprl_simple_term opid params
       else term
    end


(* ------ FREE VARS and SUBSTITUTIONS ------ *)

structure VARS = BinarySetFn(type ord_key = string val compare = String.compare)

fun remove_list vars lst =
    foldr (fn (var, vars) => VARS.delete (vars, var) handle _ => vars)
	  vars
	  lst

fun new_free_var frees var =
    let val var' = var ^ "'"
    in if VARS.member (frees, var')
       then new_free_var frees var'
       else var'
    end

fun fo_free_vars (term as TERM (operator, bterms)) =
    if is_nuprl_variable_term term
    then VARS.singleton (dest_variable term)
    else List.foldr (fn (bterm, vars) =>
			VARS.union (vars, fo_free_vars_bterm bterm))
		    VARS.empty
		    bterms

and fo_free_vars_bterm (B_TERM (vars, term)) =
    remove_list (fo_free_vars term) vars

val free_vars = fo_free_vars

structure SUB = BinaryMapFn(type ord_key = string val compare = String.compare)

datatype ran_sub = FO of nuprl_term
		 | SO of variable list * nuprl_term

fun range_sub sub =
    map (fn FO t => ([], t)
	  | SO (vs, t) => (vs, t))
	(SUB.listItems sub)

fun isin_var_list var vars = List.exists (fn v : variable => v = var) vars

fun filter_sub vars = SUB.filteri (fn (v, t) => not (isin_var_list v vars))

fun insert_sub sub (v, t) = SUB.insert (sub, v, FO t)

fun insert_ren ren (v, t) = SUB.insert (ren, v, t)

fun insert_list_sub sub list = foldr (fn (x, sub) => insert_sub sub x) sub list

fun insert_list_ren ren list = foldr (fn (x, ren) => insert_ren ren x) ren list

val empty_sub = SUB.empty : ran_sub SUB.map

val empty_ren = SUB.empty : variable SUB.map

val gen_sub = insert_list_sub empty_sub

val gen_ren = insert_list_ren empty_ren

fun apply_ren ren v t =
    case SUB.find (ren, v) of
	NONE => t
      | SOME t => t

fun fo_subst_aux sub (term as TERM (operator, bterms)) =
    if is_nuprl_variable_term term
    then let val (v, vs) = dest_so_variable term
	 in apply_sub sub (v, vs) term
	 end
    else TERM (operator, map (fo_subst_bterm sub) bterms)

and fo_subst_bterm sub (B_TERM (vars, term)) =
    let val (sub'', vars', term') = fo_rename sub (vars, term)
    in B_TERM (vars', term')
    end

and apply_sub sub (v, vs) t =
    case SUB.find (sub, v) of
	NONE => t
      | SOME (FO t) => t
      | SOME (SO (vars, t)) =>
	let val pairs = ListPair.zipEq (vars, vs)
	    val lst = map (fn (v1, v2) => (v1, mk_nuprl_variable_term v2)) pairs
	in fo_subst_aux (gen_sub lst) t
	end handle _ => raise Fail "apply_sub"

(* renames the variables in vars (and term) that occur in the
 * range of sub, and also remove the part of sub such that its
 * domain is in vars. *)
and fo_rename sub (vars, term) =
    let val sub' = filter_sub vars sub
	(* freesSub is the free variable list of sub's range.
	 * In the (SO (vars, t)) case, we only need to get the free variables
	 * in t that are not in vars. *)
	val freesSub = foldr (fn ((vs, t), vars) =>
				 let val vars' = remove_list (fo_free_vars t) vs
				 in VARS.union (vars, vars')
				 end)
			     VARS.empty
			     (range_sub sub)
	val freesTerm = fo_free_vars term
	val getNewFreeVar = new_free_var (VARS.union (freesSub, freesTerm))
	val (vars', sub'') =
	    List.foldr (fn (var, (vars, sub)) =>
			   if VARS.member (freesSub, var)
			   (* then the bound variable would capture one of
			    * the free variables of the substitution. *)
			   then let val var' = getNewFreeVar var
				    val tvar' = mk_nuprl_variable_term var'
				in (var' :: vars, insert_sub sub (var, tvar'))
				end
			   else (var :: vars, sub))
		       ([], sub')
		       vars
    in (sub'', vars', fo_subst_aux sub'' term)
    end

fun fo_subst list term = fo_subst_aux (gen_sub list) term

fun correct_lhs (vars, term) =
    if List.null vars
    then if is_nuprl_variable_term term
	 then dest_variable term
	 else raise Fail "correct_lhs"
    else let val (v, vs) = dest_so_variable term
	 in if List.all (op =) (ListPair.zipEq (vars, vs))
	     then v
	     else raise Fail "correct_lhs"
	 end handle _ => raise Fail "correct_lhs"

fun matching term1 (*term*) term2 (*lhs*) =
    let val ((opid1, params1), bterms1) = dest_term term1
	val ((opid2, params2), bterms2) = dest_term term2
    in if opid1 = opid2 andalso length bterms1 = length bterms2
       then foldr (fn ((bterm1, bterm2), sub) =>
		      SUB.insert (sub, correct_lhs bterm2, SO bterm1))
		  empty_sub
		  (ListPair.zip (bterms1, bterms2))
       else raise Fail "matching"
    end

fun unfold_ab term lhs rhs =
    let val sub = matching term lhs
    in fo_subst_aux sub rhs
    end

fun alpha_equal_bterms_aux ren (vars1,term1) (vars2,term2) =
    length vars1 = length vars2
    andalso
    let val list = ListPair.zip (vars1,vars2)
	val ren' = insert_list_ren ren list
    in alpha_equal_terms_aux ren' term1 term2
    end

and alpha_equal_terms_aux ren term1 term2 =
    (is_nuprl_variable_term term1
     andalso
     is_nuprl_variable_term term2
     andalso
     let val v1 = dest_variable term1
	 val v2 = dest_variable term2
     in v1 = apply_ren ren v2 v2
     end)
    orelse
    (let val (opr1,bterms1) = dest_term term1
	 val (opr2,bterms2) = dest_term term2
     in opr1 = opr2
	andalso
	length bterms1 = length bterms2
	andalso
	List.all (fn (bterm1, bterm2) => alpha_equal_bterms_aux ren bterm1 bterm2)
		 (ListPair.zip (bterms1, bterms2))
     end)

val alpha_equal_terms = alpha_equal_terms_aux empty_ren


(* ------ NUPRL TO EventML ------ *)

datatype alpha = NEXT of string * (unit -> alpha)

val lstAlpha =
    ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m",
     "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z"]

fun streamId [] pref () = streamId lstAlpha (pref ^ "a") ()
  | streamId (x :: xs) pref () =
    let val f = streamId xs pref
    in NEXT (pref ^ x, f)
    end

fun is_list (TERM ((("cons", _), []), [B_TERM ([], x), B_TERM ([], xs)])) =
    Option.map (fn lst => x :: lst) (is_list xs)
  | is_list (TERM ((("nil", _), []), [])) = SOME []
  | is_list _ = NONE

fun isAtomic t =
    let val opid = opid_of_term t
	val lst  = ["variable", "int", "pair", "natural_number", "prop", "universe"]
    in Option.isSome (List.find (fn x => x = opid) lst)
    end

fun printUkn t n = "-"
(*(print (toStringTerm t ^ "\n"); "---" ^ Int.toString n)*)

fun nuprl2eml_term sub (TERM ((("apply", _), []), [B_TERM ([], f), B_TERM ([], a)])) =
    nuprl2eml_atm sub f ^ " " ^ nuprl2eml_atm sub a
  | nuprl2eml_term sub (TERM ((("cons", _), []), [B_TERM ([], x), B_TERM ([], xs)])) =
    (case is_list xs of
	 SOME lst => ListFormat.fmt {final = "]", fmt = nuprl2eml_term sub, init = "[", sep = "; "} (x :: lst)
       | NONE => nuprl2eml_term sub x ^ " . " ^ nuprl2eml_term sub xs)
  | nuprl2eml_term sub (TERM ((("subtract", _), []), [B_TERM ([], x), B_TERM ([], y)])) =
    nuprl2eml_term sub x ^ " - " ^ nuprl2eml_term sub y
  | nuprl2eml_term sub (TERM ((("add", _), []), [B_TERM ([], x), B_TERM ([], y)])) =
    nuprl2eml_term sub x ^ " + " ^ nuprl2eml_term sub y
  | nuprl2eml_term sub (TERM ((("pair", _), []), [B_TERM ([], x), B_TERM ([], y)])) =
    "(" ^ nuprl2eml_term sub x ^ "," ^ nuprl2eml_term sub y ^ ")"
  | nuprl2eml_term sub (TERM ((("variable", _),       [(v,vkind)]), [])) =
    (case SUB.find (sub, v) of
	 SOME t => t
       | NONE => v)
  | nuprl2eml_term sub (TERM ((("natural_number", _), [(n,nkind)]), [])) = n
  | nuprl2eml_term sub (TERM ((("token", _), [(t,tkind)]), [])) = "`" ^ t ^ "`"
  | nuprl2eml_term sub (term as TERM ((("lambda", _), []), [B_TERM ([x], f)])) =
    printUkn term 0 (*"\\" ^ x ^ ". " ^ nuprl2eml_term sub f*)
  | nuprl2eml_term sub (TERM ((("inr", _), []), [B_TERM ([], t)])) =
    if is_nuprl_term "axiom" t
    then "false" (* NOTE: Arghh, this can be bad because inr(axiom) is not actually equal to false in EML. *)
    else "inr(" ^ nuprl2eml_term sub t ^ ")"
  | nuprl2eml_term sub (TERM ((("inl", _), []), [B_TERM ([], t)])) =
    if is_nuprl_term "axiom" t
    then "true"
    else "inl(" ^ nuprl2eml_term sub t ^ ")"
  | nuprl2eml_term sub (TERM ((("minus", _), []), [B_TERM ([], t)])) =
    "~" ^ nuprl2eml_term sub t
  | nuprl2eml_term sub (TERM ((("it", _), []), [])) = "it"
  | nuprl2eml_term sub (TERM ((("int", _), []), [])) = "Int"
  | nuprl2eml_term sub (TERM ((("bool", _), []), [])) = "Bool"
  | nuprl2eml_term sub (TERM ((("atom", _), []), [])) = "Atom"
  | nuprl2eml_term sub (TERM ((("universe", _), params), [])) = "Type"
  | nuprl2eml_term sub (TERM ((("prop", _), params), [])) = "Prop"
  | nuprl2eml_term sub (TERM ((("list", _), []), [B_TERM ([], t)])) =
    nuprl2eml_term sub t ^ " List"
  | nuprl2eml_term sub (TERM ((("product", _), []), [B_TERM ([], t1), B_TERM ([""], t2)])) =
    nuprl2eml_atm sub t1 ^ " * " ^ nuprl2eml_atm sub t2
  | nuprl2eml_term sub (TERM ((("union", _), []), [B_TERM ([], t1), B_TERM ([], t2)])) =
    nuprl2eml_atm sub t1 ^ " + " ^ nuprl2eml_atm sub t2
  | nuprl2eml_term sub (TERM ((("function", _), []), [B_TERM ([], t1), B_TERM ([""], t2)])) =
    nuprl2eml_atm sub t1 ^ " -> " ^ nuprl2eml_term sub t2
  | nuprl2eml_term sub (TERM ((("nil", _), []), [])) = "[]"
  | nuprl2eml_term sub term = (*printUkn term 1*)
    let val ((opid, params), bterms) = dest_term term
    in if List.null bterms andalso List.null params
       then opid
       else printUkn term 1
    end
    (*let val ((opid, params), bterms) = dest_term term
    in if List.all (fn (vars, t) => List.null vars) bterms
       then foldl (fn ((vars, t), str) => str ^ " " ^ nuprl2eml' t)
		  opid
		  bterms
       else toStringTerm term
    end*)

and nuprl2eml_atm sub t =
    let val str = nuprl2eml_term sub t
    in if isAtomic t then str else "(" ^ str ^ ")"
    end

fun nuprl2eml_all id sub stream (term as TERM ((("all", _), []), [B_TERM ([], t1), B_TERM ([v], t2)])) =
    if is_nuprl_term "universe" t1
    then if Option.isSome (SUB.find (sub, v))
	 then nuprl2eml_all id sub stream t2
	 else let val NEXT (tv, str) = stream ()
		  val sub' = SUB.insert (sub, v, tv)
	      in nuprl2eml_all id sub' str t2
	      end
    else printUkn term 2
  | nuprl2eml_all id sub stream (term as TERM ((("uall", _), []), [B_TERM ([], t1), B_TERM ([v], t2)])) =
    if is_nuprl_term "universe" t1
    then if Option.isSome (SUB.find (sub, v))
	 then nuprl2eml_all id sub stream t2
	 else let val NEXT (tv, str) = stream ()
		  val sub' = SUB.insert (sub, v, tv)
	      in nuprl2eml_all id sub' str t2
	      end
    else printUkn term 3
  | nuprl2eml_all id sub stream (term as TERM ((("member", _), []), [B_TERM ([], t1), B_TERM ([], t2)])) =
    if is_nuprl_term id t2
    then nuprl2eml_term sub t1
    else printUkn term 4
  | nuprl2eml_all id sub stream term = printUkn term 5

fun nuprl2eml_wf id (TERM ((("!theorem", _), [name]), [B_TERM ([], t)])) =
    nuprl2eml_all id
		  (SUB.empty : string SUB.map)
		  (streamId lstAlpha "'")
		  t
  | nuprl2eml_wf id term = printUkn term 6

fun  nuprl2eml_abs id (term as TERM ((("!abstraction", _), []), [B_TERM ([], cond), B_TERM ([], t1), B_TERM ([], t2)])) =
     if is_nuprl_term id t1
     then nuprl2eml_term SUB.empty t2
     else printUkn term 7
   | nuprl2eml_abs id term = printUkn term 8

(* nuprl2eml_wf "it" *)

end
